import sys
import traceback
from PySide6 import QtCore, QtWidgets, QtGui
import time
import asyncio
from pyvol_terminal import workers
from pyvol_terminal.settings import widgets
from pyvol_terminal.axis import axis_utils as axis_utils
import time

QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_ShareOpenGLContexts)
QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps)


windows = []


def open_new_window(interface_name, interface, config, main_window_collapse_func):
    interface_configs = config["interface_configs"]
    config["interface_configs"] = {interface_name : interface_configs[interface_name]}
    config["main_window_flag"]=False
    config["main_window_collapse_func"]=main_window_collapse_func
    w = SingeWindow(interface_name, interface, **config)
    w.show()
    windows.append(w)

class SingeWindow(QtWidgets.QMainWindow):
    def __init__(self, interface_name, interface, **config):
        super().__init__()
        self.interface_name=interface_name
        self.interface = interface  
        callback_dict = {interface_name : self.null_method} 
        main_window_flag=False
        
        self.settings_main = widgets.WindowSettings(self,
                                                  callback_dict,
                                                  main_window_flag=main_window_flag,
                                                  configs=config,
                                                  open_new_window=config["open_new_window"],
                                                  main_window_collapse_func=config["main_window_collapse_func"])
        self.initUI()
        self.showMaximized()
        
    def initUI(self):        
        self.setWindowTitle('PyVol Surface')
        self.widget_main = QtWidgets.QWidget()
        self.layout_main = QtWidgets.QVBoxLayout()
        self.setCentralWidget(self.widget_main)

        self.widget_main.setLayout(self.layout_main)
        self.layout_main.addWidget(self.settings_main)
        
        self.layout_interfaces = QtWidgets.QStackedLayout()
        self.layout_main.addLayout(self.layout_interfaces)
        
        self.layout_interfaces.addWidget(self.interface)
        self.layout_interfaces.setCurrentIndex(0)        
        
        
    def null_method(self):
        pass

    def create_interface_switch_callback_dict(self, interfaces):
        callback_dict = {}
        for interface_name, interface in interfaces.items():
            callback_dict[interface_name] = [self.interface_switch]
        return callback_dict    


class MainWindow(QtWidgets.QMainWindow):
    def __init__(self, **config):
        super().__init__()
        instrument_manager=config["instrument_manager"]
        interface_configs = config["interface_configs"]
        self.market_data_worker = config["market_data_worker"]
        self.price_process_worker = config["price_process_worker"]
        self.timer_update_plot = config["plotting_config"]["timer_update_plot"]
        
        self.last_plot_update = time.time()
        self.interfaces={}
        self.worker_slots={}
        main_window_collapse_func = config.get("main_window_collapse_func", None)

        
        interface_classes={}
        
        for idx, (name, inner_dict) in enumerate(interface_configs.items()):
            interface_classes[name] = inner_dict["interface_class"]
            if idx == 0:
                self.current_interface = name
                self.current_interface_name = name
        
        self.initInterfaces(interface_configs, instrument_manager)
        callback_dict = self.create_interface_switch_callback_dict(self.interfaces)
        for interface_name, interface in self.interfaces.items():
            self.worker_slots[interface_name] = interface.update_interface
        
        self.option_name_slot_map = {"Create New Window" : self.app.open}
        
        
        self.settings_main = widgets.WindowSettings(self,
                                                    self.interfaces,
                                                    configs=config,
                                                    open_new_window=config["open_new_window"],
                                                    main_window_collapse_func=main_window_collapse_func
                                                    )
        
        self.price_process_worker.update_interface_signal.connect(self.current_interface.update_interface)
        
        self.initUI()
        self.showMaximized()

    def create_interface_switch_callback_dict(self, interfaces):
        callback_dict = {}
        for interface_name, interface in interfaces.items():
            callback_dict[interface_name] = [self.interface_switch]
        return callback_dict    
    
    def initInterfaces(self, interface_configs, instrument_manager):
        self.interfaces={}
        self.idx_to_interface_name = {}
        self.interface_name_to_idx = {}
        self.current_interface_idx = 0
        for idx, (interface_name, interface_config) in enumerate(interface_configs.items()):
            interface_class = interface_config["interface_class"]
            interface = interface_class(instrument_manager=instrument_manager, **interface_config)
            self.interfaces[interface_name] = interface
            
            if interface_name == self.current_interface_name:
                self.current_interface = interface

            self.idx_to_interface_name[idx] = interface_name
            self.interface_name_to_idx[interface_name] = idx

    def initUI(self):        
        self.setWindowTitle('PyVol Surface')
        self.widget_main = QtWidgets.QWidget()
        self.layout_main = QtWidgets.QVBoxLayout()
        self.setCentralWidget(self.widget_main)

        self.widget_main.setLayout(self.layout_main)
        self.layout_main.addWidget(self.settings_main)
        
        self.layout_interfaces = QtWidgets.QStackedLayout()
        self.layout_main.addLayout(self.layout_interfaces)
        
        for interface in self.interfaces.values():
            self.layout_interfaces.addWidget(interface)
        
        self.layout_interfaces.setCurrentIndex(self.current_interface_idx)        
        

    def interface_switch(self, interface_name):
        prev_interface = self.current_interface
        
        self.current_interface_idx = self.interface_name_to_idx[interface_name]
        self.current_interface = self.interfaces[interface_name]
        self.layout_interfaces.setCurrentIndex(self.current_interface_idx)
        self.price_process_worker.update_interface_signal.disconnect(prev_interface.update_interface)
        self.price_process_worker.update_interface_signal.connect(self.current_interface.update_interface)

    def closeEvent(self, event):
        self.price_process_worker.update_interface_signal.disconnect(self.current_interface.update_interface)
        #loop = asyncio.get_event_loop()
        #loop.run_until_complete(loop.shutdown_asyncgens()) 
        event.accept()
    
    
def display(**config):
    app = QtWidgets.QApplication(sys.argv)
    
    websocket_config=config["websocket_config"]
    instrument_manager=config["instrument_manager"]
    data_processing_config=config["data_processing_config"]
    config["open_new_window"]=open_new_window

    price_process_worker = workers.PriceProcessor(instrument_manager,
                                                  data_processing_config=data_processing_config,
                                                  timer_update_plot=config["plotting_config"]["timer_update_plot"]
                                                  )
    market_data_worker = workers.WebsocketWorker(**websocket_config)
    market_data_worker.update_signal.connect(price_process_worker.update)
  
    config["market_data_worker"]=market_data_worker

    config["main_window_flag"]=True
    config["price_process_worker"] = price_process_worker
    
    global_font = QtGui.QFont("Neue Haas Grotesk")
    app.setFont(global_font)
    mainWin = MainWindow(**config)
    mainWin.show()
    windows.append(mainWin)
    market_data_worker.start()
    
    sys.exit(app.exec())
    


#%%
from pyqtgraph.Qt import QtCore, QtGui, QtWidgets, isQObjectAlive

from pyvol_terminal.gl_3D_graphing import meta

class AbstractGLGraphicsItem(QtCore.QObject, metaclass=meta.QABCMeta):
    _nextId = 0
    def __init__(self, parentItem: 'AbstractGLGraphicsItem' = None, **kwargs):
        super().__init__()
        self._id = AbstractGLGraphicsItem._nextId
        AbstractGLGraphicsItem._nextId += 1
        self.__blockUpdates: bool=False
        self.blockUpdates(False)
        
    def blockUpdates(self, flag):
        print(f"\nblockUpdates: {flag}")
        self.__blockUpdates=flag
        
        
    def updateBlocked(self):
        return self.__blockUpdates

    def AnotherFunc(self):
        print(f"AnotherFunc: {self.updateBlocked()}")

class ChildGLGraphicsItem(AbstractGLGraphicsItem):
    def __init__(self, parentItem: 'AbstractGLGraphicsItem' = None, **kwargs):
        super().__init__(parentItem=parentItem, **kwargs)

    def SomeFunc(self):
        self.blockUpdates(True)
        self.AnotherFunc()


child = ChildGLGraphicsItem()

child.SomeFunc()

print(f"child.updateBlocked: {child.updateBlocked()}")



from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from data_classes.classes import BaseDomain, Domain, Surface, Points
    from instruments.instruments import Option, Spot
    from instruments.utils import InstrumentManager
    from ...quantities.engines import TickEngineManager
    from ...data_classes.classes import AbstractDataClass, VolatilityData, VolVector, Surface, Points, Slice
    from ...engines.interpolation_engines import Abstract3DInterpolator
    
from PySide6 import QtWidgets
from pyvol_terminal.axis import axis_utils as utils_axis
from . import extra_widgets, stylesheets as vol_table_utils
import numpy as np
from PySide6.QtCore import Qt
from . import settings
import time
from pyvol_terminal.data_classes import builders as builders_data_classes
from ..option_monitor.extra_widgets import CustomDelegate, OptionExpiryCellItem

class MainView(QtWidgets.QTableWidget):
    def __init__(self,
                 instrument_manager: InstrumentManager,
                 tick_engine_manager: TickEngineManager=None,
                 queue_interaction: bool=False,
                 colour_styles_config: Dict[str, str]={},
                 **kwargs
                 ):
        super().__init__(**kwargs)
        self.px_type = "mid"
        self.instrument_manager=instrument_manager
        self.tick_engine_manager=tick_engine_manager

        self.volatility_data_container: Dict[str, VolatilityData] = kwargs.get("volatility_data_container", None)
        self.queue_interaction=queue_interaction

        self.ncols = 10

    
        self.metric_axis_engine=utils_axis.MetricAxisEngine(base_domain.strike,
                                                            base_domain.expiry,
                                                            "Strike",
                                                            "Expiry",
                                                            "Implied Volatility")
        self.data_container = builders_data_classes.create_surface_dataclasses(instrument_manager.config["options"]["price_types"], 
                                                                               instrument_manager.options_instrument_container.original_data,
                                                                               domain,
                                                                               self.metric_axis_engine,
                                                                               instrument_manager.options_instrument_container.get_objects(),
                                                                               interpolation_config)        

        self._initLayout(base_domain, domain, self.data_container)
        self._initCellItems()

    def _initLayout(self, base_domain: BaseDomain, domain: Domain, data_container):
        self.domain = domain
        
        
        self.setRowCount(self.domain.y_vect.size)
        self.setColumnCount(self.ncols)
      #  self.setItemDelegate(CustomDelegate(self))
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self.horizontalHeader().setSectionsMovable(False)
        self.verticalHeader().setVisible(True)
        self.verticalHeader().setSectionsMovable(False)
        self.setStyleSheet(vol_table_utils.get_settings_stylesheets("QTableWidget"))
        
        
        self.setShowGrid(False)

    def _initCellItems(self):
        
        row_values = self.domain.y_vect
        column_values = self.domain.x_vect

        row_values = self.metric_axis_engine.transform_axis("y", row_values)
        column_values = self.metric_axis_engine.transform_axis("x", column_values)
        
        self.data_container[self.px_type]["surface"].set_interpolation_domain(column_values, row_values)
        self.row_values = row_values
        self.column_values = np.linspace(column_values.min(), column_values.max(), self.ncols)
        dataclass = self.data_container[self.px_type]["surface"]
        raw = self.data_container[self.px_type]["raw"]
        
        x_metric, y_metric, z_metric, _ = self.metric_axis_engine.transform_values(self.metric_axis_engine.x_base,
                                                                                   self.metric_axis_engine.y_base,
                                                                                   **raw.getDataKwargs())
        
        dataclass.update_data(x_metric, y_metric, z_metric)
        
        self.row_labels = self.tick_engine_manager.y_engine.function(self.row_values)
        self.column_labels=self.tick_engine_manager.x_engine.function(self.column_values)
        
        for idx, label_row in enumerate(self.row_values):
            text_item = OptionExpiryCellItem(label_row)
            self.setVerticalHeaderItem(idx, text_item)
            
        for jdx, label_col in enumerate(self.column_labels):
            text_item = extra_widgets.TableColumnItem(label_col)
            self.setHorizontalHeaderItem(jdx, text_item)
        
        for idx, label_row in enumerate(self.row_values):            
            for jdx, label_col in enumerate(self.column_values):
                
                if dataclass.valid_values:
                    value = dataclass.interpolator.evaluate([label_col], [label_row])
                    new_val_str = self.tick_engine_manager.z_engine.function([value])[0]

                else:
                    new_val_str = ""                
                item = extra_widgets.OptionMetricCellItem(new_val_str)
                self.setItem(idx, jdx, item)


    def update_view(self):
        self.update_data()
        self.update_table()

    def update_data(self):
        dataclass = self.data_container[self.px_type]["surface"]
        raw = self.data_container[self.px_type]["raw"]
        x_metric, y_metric, z_metric, _ = self.metric_axis_engine.transform_values(self.metric_axis_engine.x_base,
                                                                                   self.metric_axis_engine.y_base,
                                                                                   **raw.getDataKwargs())
        dataclass.update_data(x_metric, y_metric, z_metric)
        

    def update_table(self):    
    #    self.setItemDelegate(None)
        if self.data_container[self.px_type]["surface"].valid_values:
            self.blockSignals(True)
            self.setUpdatesEnabled(False)        
            for idx, value_row in enumerate(self.row_values):
                for jdx, value_col in enumerate(self.column_values):
                    value = self.data_container[self.px_type]["surface"].interpolator.evaluate([value_col], [value_row])
                    new_val_str = self.tick_engine_manager.z_engine.function([value])[0]
                    item = self.item(idx, jdx)
                    item.setText(new_val_str)
        #    self.setItemDelegate(CustomDelegate(self))
            self.blockSignals(False)
            self.setUpdatesEnabled(True)
    
    def interface_switch(self):
        pass
    
    def switch_column_metric(self, *args):
        dataclass = self.data_container[self.px_type]["surface"]
        raw = self.data_container[self.px_type]["raw"]
        x_metric, y_metric, z_metric, _ = self.metric_axis_engine.transform_values(self.metric_axis_engine.x_base,
                                                                                   self.metric_axis_engine.y_base,
                                                                                   **raw.getDataKwargs())
        
        dataclass.update_data(x_metric, y_metric, z_metric)
        self.column_values = np.linspace(x_metric.min(), x_metric.max(), self.ncols)
        self.column_labels = self.tick_engine_manager.x_engine.function(self.column_values)
        
        for jdx, label_col in enumerate(self.column_labels):
            self.horizontalHeaderItem(jdx).setText(label_col)
        
        self.update_table()

    def table_domain(self):
        raw = self.data_container[self.px_type]["raw"]

        x_metric, y_metric, z_metric, _ = self.metric_axis_engine.transform_values(self.metric_axis_engine.x_base,
                                                                                   self.metric_axis_engine.y_base,
                                                                                   **raw.getDataKwargs())
        
        
    def switch_price_type(self, px_type):
        self.px_type=px_type
        self.update_data()
        self.update_table()



from PySide6 import QtWidgets, QtGui
import numpy as np, sys
from pyqtgraph.opengl import GLLinePlotItem, GLViewWidget, GLTextItem
from pyqtgraph.opengl.GLGraphicsItem import GLGraphicsItem

class BlockUpdateGLLinePlot(GLLinePlotItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)   

class NestedLineItem(GLLinePlotItem):
    def __init__(self, *args, **kwargs):
        self._otherNests=[]
        self._lineItems=[]
        self.blockUpdates=False
        self.blockPaint=False
        super().__init__(*args, **kwargs)   

    def addNest(self, nestedLineItem):
        self._otherNests.append(nestedLineItem)
    
    def addLineItem(self, lineItem):
        self._lineItems.append(lineItem)
        
    def paint(self):
        if not self.blockPaint:
            combined_pos = None
            for lineItem in self._lineItems:
                pos = self._mapLineItem(lineItem.pos.copy(), lineItem.transform())
                combined_pos = pos if combined_pos is None else np.vstack((combined_pos, pos))
            
            for nestedLineItem in self._otherNests:
                for lineItem in nestedLineItem._lineItems:
                    pos = self._mapLineItem(lineItem.pos.copy(), lineItem.transform())
                    combined_pos = pos if combined_pos is None else np.vstack((combined_pos, pos))
            
                self.last_child2pos = pos[-1] # last position of the child2 GLLinePlotItem's
            
            self.blockUpdates=True
            self.setData(pos=combined_pos)
            self.blockUpdates=True
            super().paint()

    def _mapLineItem(self, pos, tr):        
        if not tr.isIdentity():
            points_h = np.hstack([pos, np.ones((pos.shape[0], 1))])
            tr_matrix = np.array(tr.data()).reshape(4, 4).T
            transformed_h = points_h @ tr_matrix.T
            pos = transformed_h[:, :3]
        return pos
    
    def update(self):
        if not self.blockUpdates:
            super().update()    
    
class Window(QtWidgets.QMainWindow):
    def __init__(self, *args, app=None, **kwargs):
        super().__init__(*args, **kwargs)   
        self.gl_widget = GLViewWidget()
        self.setCentralWidget(self.gl_widget)
        
        self.gP1, self.gP2, self.gP22 = GLGraphicsItem(), GLGraphicsItem(), GLGraphicsItem()        
        #self.gP1.translate(0, 1, 1), self.gP2.translate(1, 0, -1), self.gP22.translate(1, 0, -1)

        self.nestPar1, self.nestPar2, self.nestPar22 = NestedLineItem(color="red", mode="lines"), NestedLineItem(color="yellow", mode="lines"), NestedLineItem(color="orange", mode="lines")
        #self.nestPar1.translate(0.2, 1, 1), self.nestPar2.translate(-0.2, 0, 0.2), self.nestPar22.translate(-0.2, 0, 0.2)
        
        self.nestPar1.setParentItem(self.gP1), self.nestPar2.setParentItem(self.gP2), self.nestPar22.setParentItem(self.gP22)
        self.nestPar1.translate(1, 1, 1)
        lineLength = 1
        
        for z in np.linspace(0, 10, 11):
            z /=10
            
            pos1 = [[1, 1, z], [1+lineLength, 1, z]]
            pos2 = [[1, 1, z + 0.03 ], [1+lineLength, 1, z + 0.03]]  
            
            pos22 = [[0, 1, z + 0.03], [lineLength, 1, z + 0.03]] # small offset
            l1, l2, l22 = GLLinePlotItem(pos=pos1, mode="lines", color="red"), GLLinePlotItem(pos=pos2, mode="lines"), GLLinePlotItem(pos=pos22, mode="lines")
            
            
            
            translation = -0.5, -1, 0
            
            l1.setParentItem(self.nestPar1)
            l2.setParentItem(l1)
            
            
          #  l1.translate(*translation, True), l2.translate(*translation, True), l22.translate(*translation, True)
     #       self.nestPar1.addLineItem(l1), self.nestPar2.addLineItem(l2), self.nestPar22.addLineItem(l22)
        
        self.nestPar2.blockPaint=True
      #  self.nestPar1.addNest(self.nestPar2)
        
        self.gl_widget.addItem(self.gP1)
     #   self.gl_widget.addItem(self.gP22)
        
        self.showMaximized()
        self.nestPar1.paint()
        
        for i in range(1, 11):
            txt = GLTextItem(pos=(i / 10, 0, 0), text=str(i))
            self.gl_widget.addItem(txt)
            txt = GLTextItem(pos=(0, i / 10, 0), text=str(i))
            self.gl_widget.addItem(txt)
            txt = GLTextItem(pos=( 0, 0, i / 10), text=str(i))
            self.gl_widget.addItem(txt)

        
        
    #    text2 = GLTextItem(pos=self.nestPar1.last_child2pos, text="mapping of child2 (incorrect)", font=QtGui.QFont("Arial", 6))
     #   text2.setTransform(self.nestPar1.transform())
     #   text2.setParentItem(self.gP1)
        
        text22 = GLTextItem(pos=pos22[-1], text="Child2 should be here", font=QtGui.QFont("Arial", 6))
        text22.translate(*translation, True)
      #  text22.setParentItem(self.gP22)
        
        
if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    win = Window()
    sys.exit(app.exec())


from typing import Dict, List, Callable, Union, Any
from functools import partial
from PySide6 import QtWidgets, QtGui
import sys

class RefSlotsMenu(QtWidgets.QMenu):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._connected_slots: Dict[str, List[Dict]] = {}
        self._action_objects: Dict[str, QtGui.QAction] = {}
        
    def addAction(self, action: QtGui.QAction, slots: List[Dict[str, Any]]) -> None:
        """
        Store slot configurations for an action WITHOUT connecting them.
        (Assumes connections are already handled externally)
        
        Args:
            action: The QAction to add
            slots: List of slot configurations (just stores references):
                'slot': The callable
                'connection_method': Signal type used (e.g., 'triggered')
                'args': Fixed positional args (if any)
                'kwargs': Fixed keyword args (if any)
        """
        self._action_objects[action.text()] = action
        self._connected_slots[action.text()] = slots  # Just store the configs
        super().addAction(action)  # Only adds the action to the menu
    
    def _connect_slots(self, action: QtGui.QAction, slot_configs: List[Dict]) -> None:
        for config in slot_configs:
            slot = config['slot']
            method = config['connection_method']
            args = config['args']
            kwargs = config['kwargs']
            
            signal = getattr(action, method, None)
            if signal is None:
                raise AttributeError(f"Action has no signal '{method}'")
            
            if args or kwargs:
                signal.connect(partial(slot, *args, **kwargs))
            else:
                signal.connect(slot)
    def map_slots_to_new_action(self, source_action: QtGui.QAction, target_action: QtGui.QAction) -> None:
        """
        Copy all stored slot configurations from source_action to target_action.
        Disconnects any existing connections on target_action first.
        
        Args:
            source_action: The action to copy slot configurations from
            target_action: The action to apply the slot configurations to
        """
        # Get source action text (used as key in our storage)
        
        super().addAction(source_action)
        
        source_text = source_action.text()
        
        if source_text not in self._connected_slots:
            raise KeyError(f"No slot configurations stored for action '{source_text}'")
        
        # First disconnect all existing connections on target action
        
        # Get all slot configs from source action
        slot_configs = self._connected_slots[source_text]
        
        # Connect each slot configuration to the target action
        for config in slot_configs:
            slot = config['slot']
            method = config['connection_method']
            args = config.get('args', ())
            kwargs = config.get('kwargs', {})
            
            signal = getattr(target_action, method, None)
            if signal is None:
                raise AttributeError(f"Target action has no signal '{method}'")
            
            if args or kwargs:
                signal.connect(partial(slot, *args, **kwargs))
            else:
                signal.connect(slot)
        
        # Store the configurations with the target action's text as key
        target_text = target_action.text()
        self._connected_slots[target_text] = slot_configs.copy()
        self._action_objects[target_text] = target_action
        
        self.addAction(source_action, target_action)
        
        
        
    def reconnect_slots(self, action_text: str, new_signal: str) -> None:
        if action_text not in self._connected_slots:
            raise KeyError(f"No action with text '{action_text}' found")
            
        action = self._action_objects[action_text]
        slot_configs = self._connected_slots[action_text]
        
        # First disconnect all existing connections
        for config in slot_configs:
            old_signal = getattr(action, config['connection_method'])
            try:
                old_signal.disconnect(config['slot'])
            except TypeError:
                pass
        
        # Update connection method in all configs
        for config in slot_configs:
            config['connection_method'] = new_signal
        
        # Reconnect with new signal
        self._connect_slots(action, slot_configs)
    
    def get_action(self, text: str) -> QtGui.QAction:
        return self._action_objects.get(text)
    
    def get_slot_configs(self, action_text: str) -> List[Dict]:
        return self._connected_slots.get(action_text, [])
            
    def add_signal_to_existing_slots(self, action_text: str, new_signal_name: str) -> None:
        """
        Connect all existing slots of an action to a new signal
        
        Args:
            action_text: Text of the action to modify
            new_signal_name: Name of the new signal to connect to (e.g. 'triggered', 'toggled')
        """
        if action_text not in self._connected_slots:
            raise KeyError(f"No action with text '{action_text}' found")
            
        action = self._action_objects[action_text]
        slot_configs = self._connected_slots[action_text]
        
        # First, disconnect all slots from their current signals
        for config in slot_configs:
            old_signal = getattr(action, config['connection_method'])
            slot = config['slot']
            args = config.get('args', ())
            kwargs = config.get('kwargs', {})
            
            try:
                if args or kwargs:
                    # Disconnect the partial version
                    old_signal.disconnect(partial(slot, *args, **kwargs))
                else:
                    # Disconnect the direct slot
                    old_signal.disconnect(slot)
            except (TypeError, RuntimeError):
                # Ignore if not connected
                pass
        
        # Get the new signal
        new_signal = getattr(action, new_signal_name, None)
        if new_signal is None:
            raise AttributeError(f"Action has no signal '{new_signal_name}'")
        
        # Connect all existing slots to the new signal
        for config in slot_configs:
            slot = config['slot']
            args = config.get('args', ())
            kwargs = config.get('kwargs', {})
            
            if args or kwargs:
                new_signal.connect(partial(slot, *args, **kwargs))
            else:
                new_signal.connect(slot)
        
        # Update the connection method in config for future reference
        for config in slot_configs:
            config['connection_method'] = new_signal_name
                
                
# Test handlers
def handler1(checked, name):
    print(f"Handler1: {name} is {checked}")

def handler2(checked):
    print(f"Handler2: {checked}")

def handler3(checked, extra_arg, kwarg=None):
    print(f"Handler3: Checked={checked}, Extra={extra_arg}, KW={kwarg}")


def basic_handler(checked):
    print(f"Basic handler - Checked: {checked}")

def named_handler(checked, name):
    print(f"Named handler - {name} is {checked}")

def partial_handler(checked, fixed_arg, kwarg=None):
    print(f"Partial handler - Checked: {checked}, Fixed: {fixed_arg}, KW: {kwarg}")

# Create application
app = QtWidgets.QApplication(sys.argv)
# Create main window with close button
window = QtWidgets.QMainWindow()
window.setWindowTitle("Menu Close Example")
window.resize(300, 200)
# Create our custom menu
menu = RefSlotsMenu("Demo Menu")
# Existing action with slots connected to 'triggered'
action = QtGui.QAction("Test Action", menu)
action.setCheckable(True)  # Required for 'toggled' signal
menu.addAction(action, [
    {'slot': handler1, 'connection_method': 'triggered', 'args': (True,)},
    {'slot': handler2, 'connection_method': 'triggered'}
])

# Replace 'triggered' with 'toggled' (disconnects old signals automatically)
menu.add_signal_to_existing_slots("Test Action", "toggled")
# Add the target action to the menu

tool_button = QtWidgets.QToolButton()
tool_button.setText("Show Menu")
tool_button.setPopupMode(QtWidgets.QToolButton.InstantPopup)
tool_button.setMenu(menu)

# Add button to toolbar
toolbar = window.addToolBar("Main Toolbar")
toolbar.addWidget(tool_button)

# Show window
window.show()

# Execute application
sys.exit(app.exec())


import numpy as np
from OpenGL import GL as opengl
from OpenGL import GLU

def map_2D_coords_to_3D(widget, x, y):
    widget_width = widget.width()
    widget_height = widget.height()
    device_pixel_ratio = widget.window().screen().devicePixelRatio()

    x_norm = x / widget_width
    y_norm = y / widget_height

    viewport = opengl.glGetIntegerv(opengl.GL_VIEWPORT)
    _, _, viewport_width, viewport_height = viewport
    
    mouse_x_physical = x_norm * viewport_width
    mouse_y_physical = y_norm * viewport_height
    mouse_y_physical = viewport_height - mouse_y_physical 
    
    depth = opengl.glReadPixels(int(mouse_x_physical), int(mouse_y_physical), 1, 1, opengl.GL_DEPTH_COMPONENT, opengl.GL_FLOAT)[0][0]

    modelview = np.array(widget.viewMatrix().data()).reshape(4, 4)
    projection = np.array(widget.projectionMatrix(viewport,
                                                  (0, 0, device_pixel_ratio * widget_width,
                                                   device_pixel_ratio* widget_height)
                                                  ).data()).reshape(4, 4)
    
    world_x, world_y, world_z = GLU.gluUnProject(mouse_x_physical, mouse_y_physical, depth, modelview, projection, viewport)

    return world_x, world_y, world_z

def map_3D_coords_to_2D(widget, world_x, world_y, world_z):
    device_pixel_ratio = widget.window().screen().devicePixelRatio()
    widget_height = widget.height()
    widget_width = widget.width()
    
    viewport = opengl.glGetIntegerv(opengl.GL_VIEWPORT)
    _, _, viewport_width, viewport_height = viewport
    
    modelview = np.array(widget.viewMatrix().data()).reshape(4, 4)
    projection = np.array(widget.projectionMatrix(viewport,
                                                  (0, 0, device_pixel_ratio * widget_width, device_pixel_ratio * widget_height)).data()).reshape(4, 4)
    
    px_x, px_y, px_z = GLU.gluProject(world_x, world_y, world_z, modelview, projection, viewport)
    px_x = viewport_width - px_x
    px_y = viewport_height - px_y
    
    return px_x, px_y, px_z




#%% 

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes

class CubeVertex:
    def __init__(self, point, perp_axis_edge):
        self.point = point
        self.perp_axis_edge = perp_axis_edge

def get_boundary_intersections(p1, p2):
    x1, y1 = p1
    x2, y2 = p2
    intersections = []
    
    if x1 == x2:
        return [(x1, -10), (x1, 10)]
    
    m = (y2 - y1)/(x2 - x1)
    b = y1 - m*x1
    
    y = m*(-10) + b
    if -10 <= y <= 10:
        intersections.append((-10, y))
        
    y = m*10 + b
    if -10 <= y <= 10:
        intersections.append((10, y))
        
    x = (-10 - b)/m
    if -10 <= x <= 10:
        intersections.append((x, -10))
        
    x = (10 - b)/m
    if -10 <= x <= 10:
        intersections.append((x, 10))
        
    return intersections

def plot_vertex_lines(ax: Axes, m: float, length: float = 0.3):
    cube_vertices = get_cube_coordinates()
    cube_vertices = [*cube_vertices[:3], cube_vertices[3]]

    for vertex in cube_vertices:
        x0, y0 = vertex.point
        if np.isinf(m):
            dx, dy = 0, length / 2
        else:
            dx = (length / 2) / np.sqrt(1 + m**2)
            dy = m * dx
        x_vals = [x0 - dx, x0 + dx]
        y_vals = [y0 - dy, y0 + dy]
        ax.plot(x_vals, y_vals, color='orange')

def get_cube_coordinates():
    p1 = CubeVertex((0, 0), (0, 1))
    p2 = CubeVertex((0, 1), (1, 0))
    p3 = CubeVertex((1, 0), (1, 1))
    p4 = CubeVertex((1, 1), (1, 0))
    
    return p1, p2, p3, p4
    
def plot_rectangle(ax: Axes, p1, p2, p3, p4, color='k', linestyle='-'):
    points = [p.point for p in [p1, p2, p4, p3, p1]]
    x_coords = [pt[0] for pt in points]
    y_coords = [pt[1] for pt in points]
    ax.plot(x_coords, y_coords, color=color, linestyle=linestyle, linewidth=1.2)

def plot_line_with_perpendiculars(ax: Axes, p1, p2, lim_ranges):
    intersections = get_boundary_intersections(p1, p2)
    
    if len(intersections) == 2:
        ax.plot([intersections[0][0], intersections[1][0]], 
                [intersections[0][1], intersections[1][1]], 'b-')
    
    dx = p2[0] - p1[0]
    dy = p2[1] - p1[1]
    perp_slope = -dx/dy if dy != 0 else float('inf')
    
    for point in [p1, p2]:
        x0, y0 = point
        intersections = []
        
        if np.isinf(perp_slope):  
            intersections.append((x0, -10))
            intersections.append((x0, 10))
        else:
            perp_b = y0 - perp_slope*x0
            
            y = perp_slope*(-10) + perp_b
            if -10 <= y <= 10:
                intersections.append((-10, y))
                
            y = perp_slope*10 + perp_b
            if -10 <= y <= 10:
                intersections.append((10, y))
                
            x = (-10 - perp_b)/perp_slope
            if -10 <= x <= 10:
                intersections.append((x, -10))
                
            x = (10 - perp_b)/perp_slope
            if -10 <= x <= 10:
                intersections.append((x, 10))
        
        if len(intersections) == 2:
            ax.plot([intersections[0][0], intersections[1][0]], 
                    [intersections[0][1], intersections[1][1]], 'r-')
    
    cube_coords = get_cube_coordinates()
    plot_rectangle(ax, *cube_coords)
    
    plot_vertex_lines(ax, perp_slope, 1)
    
    ax.plot(p1[0], p1[1], 'go', markersize=5)
    ax.plot(p2[0], p2[1], 'go', markersize=5)
    ax.set_xlim(lim_ranges[0], lim_ranges[1])
    ax.set_ylim(lim_ranges[0], lim_ranges[1])

    ax.grid(True)
    ax.set_aspect('equal', adjustable='box')

fig, axs = plt.subplots(ncols=2, nrows=2, layout='constrained')
ax1, ax2, ax3, ax4 = axs.flatten()

lim_ranges = -2, 4

p1 = (0, 1)
p2 = (1.5, -0.3)
plot_line_with_perpendiculars(ax1, p1, p2, lim_ranges)

p1 = (0, 2)
p2 = (1.5, .7)
plot_line_with_perpendiculars(ax2, p1, p2, lim_ranges)

p1 = (-0.2, 1.05)
p2 = (1.8, 0.95)
plot_line_with_perpendiculars(ax3, p1, p2, lim_ranges)

plt.show()


import sys
from PySide6.QtWidgets import (QApplication, QMainWindow, QDockWidget, QTextEdit, 
                               QMenu, QLabel, QWidget)
from PySide6.QtCore import Qt, QEvent, QPoint
from PySide6.QtGui import QContextMenuEvent


class DockWidget(QDockWidget):
    def __init__(self, title, parent=None):
        super().__init__(title, parent)
        self.setAllowedAreas(Qt.AllDockWidgetAreas)
        
        # Create content for the dock
        content = QLabel("Right-click the title bar to dock/undock this widget")
        content.setAlignment(Qt.AlignCenter)
        self.setWidget(content)
        
        # Track floating state changes
        self.topLevelChanged.connect(self.handle_floating_change)
        
        # Install event filter for both docked and floating states
        self.update_event_filter()
    
    def handle_floating_change(self, floating):
        self.update_event_filter()
    
    def update_event_filter(self):
        # Remove any existing filter
        try:
            self.title_bar.removeEventFilter(self)
        except:
            pass
        
        # Find the current title bar widget
        if self.isFloating():
            # For floating state, the title bar is the window's title bar
            self.title_bar = self.findChild(QWidget, "qt_dockwidget_floatwindow")
        else:
            # For docked state, use the standard title bar
            self.title_bar = self.findChild(QWidget, "qt_dockwidget_titlebar")
        
        if self.title_bar:
            self.title_bar.installEventFilter(self)
    
    def eventFilter(self, obj, event):
        # Handle right-click on title bar
        if (obj == self.title_bar and 
            event.type() == QEvent.MouseButtonRelease and 
            event.button() == Qt.RightButton):
            
            # For floating windows, we need to adjust the position
            if self.isFloating():
                pos = self.mapToGlobal(QPoint(event.pos().x(), self.title_bar.height()))
            else:
                pos = event.globalPos()
            
            self.show_titlebar_menu(pos)
            return True
        return super().eventFilter(obj, event)
    
    def show_titlebar_menu(self, pos):
        menu = QMenu(self)
        
        # Toggle floating state action
        if self.isFloating():
            action_text = "Dock"
        else:
            action_text = "Float"
            
        float_action = menu.addAction(action_text)
        float_action.triggered.connect(self.toggle_floating)
        
        # Add other standard actions
        close_action = menu.addAction("Close")
        close_action.triggered.connect(self.close)
        
        menu.exec(pos)
    
    def toggle_floating(self):
        self.setFloating(not self.isFloating())


class MainWindow(QMainWindow):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("DockWidget Right-Click Example")
        self.setGeometry(100, 100, 800, 600)
        
        # Central widget
        self.setCentralWidget(QTextEdit("Main Window Content"))
        
        # Create dock widget
        self.dock = DockWidget("Tool Panel", self)
        self.addDockWidget(Qt.RightDockWidgetArea, self.dock)
        
        # Make it float initially to demonstrate the feature
        self.dock.setFloating(True)


if __name__ == "__main__":
    app = QApplication(sys.argv)
    window = MainWindow()
    window.show()
    sys.exit(app.exec())


from __future__ import annotations
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING
if TYPE_CHECKING:
    from ...quantities.engines import TickEngineManager
    from .settings import Settings
    from ..custom_widgets import CustomPlotDataItem
    from .main_view import MainView
    from pyvol_terminal.interfaces.volatility_surface.pyvol_GL3DViewWidget import PyVolGL3DViewWidget
    from .interface import Interface
    
from ...gl_3D_graphing.graphics_items import GL3DAxisItem
from .axes_widgets import PyVolGL3DAxisItem
import pyqtgraph as pg
from .. import custom_widgets
from . import axis_widgets 
from PySide6 import QtCore, QtGui, QtWidgets

       
def initAxis(view_widget: PyVolGL3DViewWidget,
             tick_engine_manager: TickEngineManager,
             ) -> Tuple[List[PyVolGL3DAxisItem], List[axis_widgets.CustomAxisItem]]:

    x_label = "Strike"
    y_label = "Expiry"
    z_label = "Implied Volatility"

    x_ax_3D = [[view_widget.vb.state["worldRange"], 0, (0, 0, 0), 1],
                {"text" : x_label,
                "showValues" : True,
                "valuesColor" : "white",
                "labelColor" : "white",
                "faceTickOffset" : -0.1,
                "syncToView" : True,
                "showGrid" : True,
                "showTickMargin" : True,         
                }
              ]
    
    y_ax_3D = [[view_widget.vb.state["worldRange"], 1, (0, 0, 0), 0],
                {"text" : y_label,
                "showValues" : True,
                "valuesColor" : "yellow",
                "labelColor" : "yellow",
                "faceTickOffset" : -0.1,
                "syncToView" : True,
                "showGrid" : True,
                "showTickMargin" : True,     
                }
               ]

    z_ax_3D = [[view_widget.vb.state["worldRange"], 2, (0, 0, 0), 0],
                {"text" : z_label,
                "showValues" : True,
                "valuesColor" : "cyan",
                "labelColor" : "cyan",
                "faceTickOffset" : -0.1,
                "syncToView" : True,
                "showGrid" : True,
                "showTickMargin" : True,   
                "graphFaceOffset" : -0.3,
                }
               ]
    
    zz_ax_3D = [[view_widget.vb.state["worldRange"], 2, (1, 1, 0), 1],
                {"text" : z_label,
                "showValues" : True,
                "valuesColor" : "cyan",
                "labelColor" : "cyan",
                "faceTickOffset" : -0.1,
                "syncToView" : True,
                "showGrid" : True,
                "showTickMargin" : True,     
                "graphFaceOffset" : -0.3,
                }
               ]
    
    axes_with_ticks = [x_ax_3D, y_ax_3D, z_ax_3D, zz_ax_3D]

    axes_3D_items = [PyVolGL3DAxisItem.fromViewBoxRange(*item[0], **item[1], tick_engine=tick_engine_manager.get_engine(ax)) for ax, item in zip("xyzz", axes_with_ticks)]
    
    xn_ax_3D = [[view_widget.vb.state["worldRange"], 0, (0, 0, 0), 2],
                {"text" : z_label,
                "showValues" : False,
                "faceTickOffset" : -0.1,
                "syncToView" : True,
                "showGrid" : True,
                "showTickMargin" : True,   
                "showLabel" : False,
                "graphFaceOffset" : -0.3,
                }
                ]
    
    yn_ax_3D = [[view_widget.vb.state["worldRange"], 1, (1, 1, 0), 2],
                {"text" : z_label,
                "showValues" : False,
                "faceTickOffset" : -0.1,
                "syncToView" : True,
                "showGrid" : True,
                "showTickMargin" : True,     
                "showLabel" : False,
                "graphFaceOffset" : -0.3,
                }
                ]
    
    axes_no_ticks = [xn_ax_3D,yn_ax_3D]
    
    axes_3D_items = axes_3D_items + [PyVolGL3DAxisItem.fromViewBoxRange(*item[0], **item[1]) for item in axes_no_ticks]
    
    return axes_3D_items




import pyqtgraph.opengl as gl
from PySide6 import QtWidgets
from pyqtgraph.opengl import GLTextItem
from PySide6 import QtGui, QtCore

class RotatedGLTextItem(GLTextItem):
    def __init__(self, text=None, angle=45, color=(1,1,1,1), parent=None):
        super().__init__(text=text, color=color, parent=parent)
        self.angle = angle
        
    def paint(self):
        glPushMatrix()
        glRotatef(self.angle, 0, 0, 1)  # Rotate around Z-axis
        super().paint()
        glPopMatrix()
app = QtWidgets.QApplication([])
w = gl.GLViewWidget()
w.show()

# Add a rotated text item
text = RotatedGLTextItem("Rotated Text", angle=45)
w.addItem(text)

app.exec_()


from PySide6.QtCore import QThread, QTimer
import time

# Shared counter (accessed by multiple threads)
counter = 0

class WorkerThread(QThread):
    def run(self):
        global counter
        for _ in range(1000):
            current = counter
            time.sleep(0.001)  # Simulate processing delay
            counter = current + 1

# Create and start multiple threads
threads = []
for _ in range(5):
    thread = WorkerThread()
    thread.start()
    threads.append(thread)

# Wait for all threads to finish
for thread in threads:
    thread.wait()

print("Final counter value (expected 5000):", counter)  # Likely less than 5000 due to race condition!


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any
import numpy as np
from abc import ABC, abstractmethod
from scipy import interpolate


class Abstract3DInterpolator(ABC):
    def __init__(self, n_x, n_y):
        self._n_x, self._n_y= n_x, n_y
        self._fitted=False
    
    def grid_shape(self) -> Tuple[int, int]:
        return self._n_x, self._n_y
    
    @abstractmethod
    def fit(self,
            x: np.ndarray,
            y: np.ndarray,
            z: np.ndarray,
            *args: Tuple[Any],
            **kwargs: Dict[str, Any]
            ) -> None:
        ...
        
    @abstractmethod
    def evaluate(self,
                 xi: np.ndarray,
                 yi: np.ndarray,
                 *args: Tuple[Any],
                 **kwargs: Dict[str, Any]
                 ) -> np.ndarray:
        ...


class CustomRBFInterpolator(Abstract3DInterpolator, interpolate.RBFInterpolator):
    def __init__(self, n_x, n_y):
        Abstract3DInterpolator.__init__(self, n_x, n_y)
        
    def fit(self,
            x: np.ndarray,
            y: np.ndarray,
            z: np.ndarray, 
            args: Tuple[Any],
            kwargs: Dict[str, Any]
            ) -> None:
        x = x.flatten()
        y = y.flatten()
        z = z.flatten()
        points = np.column_stack((x, y))

        interpolate.RBFInterpolator.__init__(self, points, z, *args, **kwargs)
        self._fitted = True      
        
          
    def evaluate(self,
                 xi: np.ndarray,
                 yi: np.ndarray,
                 args: Tuple[Any],
                 kwargs: Dict[str, Any]
                 ) -> np.ndarray:
        if not self._fitted:
            raise RuntimeError("Call `fit` before `evaluate`.")

        xi, yi = np.meshgrid(xi, yi)
        points = np.column_stack((xi.flatten(), yi.flatten()))
        new_vals = self(points).reshape(*self.grid_shape())
        return new_vals
    
    
class CustomBSplineInterpolator(Abstract3DInterpolator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tck = None
    
    def fit(self,
            x: np.ndarray,
            y: np.ndarray,
            z: np.ndarray,
            args: Tuple[Any],
            kwargs: Dict[str, Any]
            ) -> None:
        kwargs["s"] = x.size
        self.tck = interpolate.bisplrep(x, y, z, *args, **kwargs) 
        self._fitted = True     
        
    def evaluate(self,
                 xi: np.ndarray,
                 yi: np.ndarray,
                 args: Tuple[Any],
                 kwargs: Dict[str, Any]
                 ) -> np.ndarray:
        if not self._fitted:
            raise RuntimeError("Call `fit` before `evaluate`.")

        x_vals = np.unique(xi)
        y_vals = np.unique(yi)
       #x_vals.sort()
        #y_vals.sort()
        return interpolate.bisplev(x_vals, y_vals, self.tck, *args, **kwargs) 


class CustomSplineInterpolator:
    def __init__(self, ):
        self.x=None
        self.y=None
        self.z=None
        self.spline=None
        self.k=2
    
    def fit(self,
            x: np.ndarray,
            y: np.ndarray,
            ) -> None:
        self.spline = interpolate.make_interp_spline(x, y, k=self.k)
        self._fitted = True     
    
    def evaluate(self, xi):
        return self.spline(xi)


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any

    
from PySide6 import QtCore, QtWidgets
from . import settings
from .. import abstract_classes
from ...quantities import engines
from ...data_classes import builders as builders_data_classes
from ..abstract_classes import ABCInterface, ABC
from ...instruments.utils import InstrumentManager
from ...workers import SurfaceCalibration
from ...quantities.engines import TickEngineManager
from . import pyvol_GL3DViewWidget
from ...data_classes.classes import AbstractDataClass, VolatilityData, VolVector, Surface, Points, Slice
from ...data_classes import builders as builders_data_classes
from PySide6 import QtCore, QtWidgets, QtGui
from . import init_helpers
from .. import custom_widgets, subplot
import pyqtgraph as pg
from . import pyvol_GL3DViewWidget
from pyvol_terminal.misc_classes import PriceText
import time
from ...gl_3D_graphing.graphics_items import GL3DGraphicsItems
from .pyvol_GL3DGraphicsItems import PyVolGL3DSurfacePlotItem, PyVolGL3DScatterPlotItem

import traceback
from ..abstract_classes import ABCInterface
from dataclasses import dataclass, field, InitVar
import numpy as np
from pprint import pprint   
from .pyvol_GL3DViewWidget import PyVolGL3DViewWidget
from .dock_window import SplitDockWidget, SubPlotItem
from PySide6.QtCore import Qt

QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_ShareOpenGLContexts)
QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps)


class Interface(ABCInterface):
    aboutToCloseSig = QtCore.Signal(QtWidgets.QMainWindow)
    
    
    def __init__(self,
                 instrument_manager: InstrumentManager=None,
                 tick_engine_manager=None,
                 vol_vect_container=None,
                 queue_interaction: bool=False,
                 colour_styles_config: Dict[str, str]={},
                 **configs
                 ):
        super().__init__()

        self.all_price_types = instrument_manager.config["options"]["price_types"]
        self.splot_flag=True
        self.subplots_flag=True
        self.on_view=False
        self.plot_interaction_buffer=[]
        self._first_show=True
        
        if tick_engine_manager is None:
            tick_engine_manager = engines.TickEngineManager("Strike",
                                                            "Date",
                                                            "Implied Volatility"
                                                            )
            
        self.volatility_data_container = builders_data_classes.create_volatility_data_from_vol_vect(vol_vect_container)

        self.pyvol_gl_widget: PyVolGL3DViewWidget=None
        
        self.queue_interaction=queue_interaction
        
        self.interact_finish_callbacks=[]
        
        self.colour_styles_config=colour_styles_config

        all_price_types = instrument_manager.config["options"]["price_types"]
        
        self._dclass_name_subplot_title_map: Dict[str, str]={"skew" : "Skew",
                                                             "term" : "Term Structure"
                                                             }
        self._all_assets = ["surface", "scatter"]
        self._displayed_assets = ["surface", "scatter"]
        self._subplot_widget_container: Dict[str, SubPlotItem]={}
        
        self.displayed_price_types=[]
        self.displayed_price_plot_type_flags = {px_type: {"surface" : False, "scatter" : False} for px_type in self.all_price_types}
        self.displayed_price_plot_type_names = {px_type: {"surface" : None, "scatter" : None} for px_type in all_price_types}
        
        self.plot_state = {"surface" : True,
                           "scatter" : True,
                           "subplots" : True,
                           }

        self.scatter_from_calibrated=True

        self._initScene(instrument_manager, tick_engine_manager, **configs)
        self.showMaximized()
        
    
    def addSettings(self, settings):
        if not settings is None:
            self.addToolBar(settings)
    
    def get_calibration_slot(self) -> Callable:
        return self.update_view

    def _initScene(self,
                   instrument_manager: InstrumentManager,
                   tick_engine_manager: TickEngineManager,
                   **configs
                   ):
        spot_objects = [item for item in instrument_manager.spot_instrument_container.objects.values()]
        spot_text = PriceText(spot_objects)
        central_widget = PyVolGL3DViewWidget(spot_text, computeNormals=False, **configs)
        self.setCentralWidget(central_widget)
        
        settings_widget = self._buildSettings(instrument_manager, tick_engine_manager)
        settings_widget.setAllowedAreas(Qt.TopToolBarArea | Qt.BottomToolBarArea)

        split_dock_widget = SplitDockWidget(tick_engine_manager=tick_engine_manager)
        split_dock_widget.setAllowedAreas(Qt.LeftDockWidgetArea | Qt.RightDockWidgetArea)

        self.addDockWidget(Qt.RightDockWidgetArea, split_dock_widget)
        self.addToolBar(settings_widget)
        
    def _buildSettings(self,
                       instrument_manager: InstrumentManager,
                       tick_engine_manager: TickEngineManager
                       ) -> settings.Settings:
        
        settings_widget = settings.Settings(self)

        
        settings_widget.add_window_menu(self.switch_axis_units)
        settings_widget.add_vol_src_window_menu(self.all_price_types, 
                                                [self.toggle_price_type, instrument_manager.update_metric_calculations]
                                                )
        
        switch_axis_slots = [tick_engine_manager.change_function]
        switch_axis_slots = switch_axis_slots + [vola_data.vol_vector.metric_engine.change_function for vola_data in self.volatility_data_container.values()]
        switch_axis_slots = switch_axis_slots + [self.switch_axis]      

        settings_widget.create_combobox_menu("Moneyness",
                                             ["Strike", "Delta", "Moneyness", "Log-Moneyness", "Standardised-Moneyness"],
                                             switch_axis_slots,
                                             "x"
                                             )

        settings_widget.create_combobox_menu("Tenor",
                                             ["Date", "Years"],
                                             switch_axis_slots,
                                             "y"
                                             )

        settings_widget.create_combobox_menu("Volatility",
                                             ["Implied Volatility", "Variance", "Total Volatility", "Total Variance"],
                                             switch_axis_slots,
                                             "z"
                                             )

        settings_widget.create_combobox_menu("Toggle Crosshairs",
                                             ["On", "Off"],
                                             self.centralWidget().toggle_crosshairs,
                                             )
        
        #settings_widget.create_vol_src_menu(self.all_price_types,
                                            #[self.toggle_price_type, instrument_manager.update_metric_calculations]
                                            #)

        settings_widget.create_toggle_buttons("Toggle 3D Assets",
                                              ["surface", "scatter"], 
                                              [self.toggle_3D_objects],
                                              )
        return settings_widget
    
    def _internal_update_view(self, calibrated_surface_container: Dict[str, SurfaceCalibration]):
            print(calibrated_surface_container)
            self.update_dataclasses(calibrated_surface_container)
            self.update_plots()     
            self._queued_update=False

    def update_dataclasses(self, calibrated_surface_container: Dict[str, SurfaceCalibration]):
        for px_type in self.displayed_price_types:
            surface_engine = calibrated_surface_container[px_type]
            volatility_data = self.volatility_data_container[px_type]
            for plot_type in self._displayed_assets:
                dataclass = volatility_data.get_dataclass(plot_type)
                dataclass.evaluate_from_engine(surface_engine)

    def update_plots(self):
        for px_type in self.displayed_price_types:
            volatility_data = self.volatility_data_container[px_type]
            for plot_type in self._displayed_assets:
                dataclass = volatility_data.get_dataclass(plot_type)
                internal_id = self.displayed_price_plot_type_names[px_type][plot_type]
                
                gl_plotdataitem = self.centralWidget().plotItem(internal_id)
                gl_plotdataitem.setData(**dataclass.plot_item_kwargs())
                                
                if plot_type == "surface":
                    for sub_dataclass in volatility_data.surface_children():
                        if sub_dataclass.valid_values:
                            subplot_title = self._dclass_name_subplot_title_map[sub_dataclass.name]
                            subplot_widget = self._subplot_widget_container[subplot_title]
                            subplot_widget.pricePlotDataItem(px_type).setData(**sub_dataclass.plot_item_kwargs())
                                                
    def _add_surface_plot_data_item(self, px_type):
        if not self.displayed_price_plot_type_flags[px_type]["surface"]: 
            volatility_data = self.volatility_data_container[px_type]            
            surface_plotdataitem = PyVolGL3DSurfacePlotItem(px_type=px_type,
                                                            colormap=self.colour_styles_config["surface"][px_type],
                                                            color=self.colour_styles_config["scatter"][px_type],
                                                            **volatility_data.surface.plot_item_kwargs()
                                                            )
            self.centralWidget().addItem(surface_plotdataitem, ignoreBounds=False)
            
            self.displayed_price_plot_type_flags[px_type]["surface"]=True
            self.displayed_price_plot_type_names[px_type]["surface"]=surface_plotdataitem.id()
            
            if self.plot_state["subplots"]:
                self._add_subplot_line(px_type, volatility_data)

    def _add_subplot_line(self, px_type: str, volatility_data: VolatilityData):
        pen = pg.mkPen(color=self.colour_styles_config["scatter"][px_type])
        for sub_dataclass in volatility_data.surface_children():
            subplot_title = self._dclass_name_subplot_title_map[sub_dataclass.name]
            if subplot_title in self._subplot_widget_container:
                subplot_widget = self._subplot_widget_container[subplot_title]
                _ = subplot_widget.plotFromDataClass(sub_dataclass,
                                                     px_type=px_type,
                                                     pen=pen
                                                     )

    def _add_scatter_plot_data_item(self, px_type):        
        if not self.displayed_price_plot_type_flags[px_type]["scatter"]:
            volatility_data = self.volatility_data_container[px_type]
            scatter_plotdataitem = PyVolGL3DScatterPlotItem(px_type=px_type,
                                                            color=self.colour_styles_config["scatter"][px_type],
                                                            **volatility_data.scatter.plot_item_kwargs()
                                                            )
            
            self.centralWidget().addItem(scatter_plotdataitem, ignoreBounds=False)
            
            self.displayed_price_plot_type_flags[px_type]["scatter"]=True
            self.displayed_price_plot_type_names[px_type]["scatter"]=scatter_plotdataitem.id()

    def _remove_surface_plot_data_item(self, px_type):
        if px_type in self.centralWidget().surface_plotitems:
            self.centralWidget().removeItem(self.centralWidget().surface_plotitems[px_type])
            self.displayed_price_plot_type_names[px_type]["surface"]=None
            self.displayed_price_plot_type_flags[px_type]["surface"]=False
            
            if self.plot_state["subplots"]:
                volatility_data = self.volatility_data_container[px_type]

                for sub_dataclass in volatility_data.surface_children():
                    subplot_title = self._dclass_name_subplot_title_map[sub_dataclass.name]
                    if subplot_title in self._subplot_widget_container:
                        subplot_widget = self._subplot_widget_container[subplot_title]
                        subplot_widget.removePricePlotDataItem(px_type)              

    def _remove_scatter_plot_data_item(self, px_type):
        if px_type in self.centralWidget().scatter_plotitems:
            self.centralWidget().removeItem(self.centralWidget().scatter_plotitems[px_type])
            self.displayed_price_plot_type_names[px_type]["scatter"]=None
            self.displayed_price_plot_type_flags[px_type]["scatter"]=False
            
    def toggle_price_type(self, flag, px_type):
        if flag:
            if not px_type in self.displayed_price_types:
                if self.plot_state["surface"]:
                    self._add_surface_plot_data_item(px_type)
                if self.plot_state["scatter"]:
                    self._add_scatter_plot_data_item(px_type)
                self.displayed_price_types.append(px_type)    
        else:
            if px_type in self.displayed_price_types:
                if self.plot_state["surface"]:
                    self._remove_surface_plot_data_item(px_type)
                if self.plot_state["scatter"]:
                    self._remove_scatter_plot_data_item(px_type)
                self.displayed_price_types.remove(px_type)

    def _toggle_surface(self, check):
        if check == self.plot_state["surface"]:
            return
        if not self.plot_state["surface"]:
            for px_type in self.centralWidget().get_displayed_price_types().copy():
                self._add_surface_plot_data_item(px_type)
        else:
            for px_type in self.centralWidget().get_displayed_price_types().copy():
                self._remove_surface_plot_data_item(px_type)
        self.plot_state["surface"] = not self.plot_state["surface"]
        
    def _toggle_scatter(self, check):
        if check == self.plot_state["scatter"]:
            return 
        if not self.plot_state["scatter"]:
            for px_type in self.centralWidget().get_displayed_price_types().copy():
                self._add_scatter_plot_data_item(px_type)
        else:
            for px_type in self.centralWidget().get_displayed_price_types().copy():
                self._remove_scatter_plot_data_item(px_type)
        self.plot_state["scatter"] = not self.plot_state["scatter"]
        
    def toggle_3D_objects(self, check, plot_type):
        if plot_type == "surface":
            self._toggle_surface(check)
        elif plot_type == "scatter":
            self._toggle_scatter(check)
            
    def interface_switch(self, *args):
        self._internal_update_view()

    def plot_queued_update(self):
        if self._queued_update:
            self.update_plots()    
            self._queued_update=False
    
    def generate_random(self):
        pass
    
    def switch_axis(self, axis_label, axis_direction):
        if isinstance(axis_direction, str):
            str_int_map = {"x" : 0,
                           "y" : 1,
                           "z" : 2,
                           }
            axis_direction=str_int_map[axis_direction]
        for axis_item in self.centralWidget().axes[axis_direction]:
            if axis_item.style["showLabel"]:
                axis_item.setLabel(axis_label)
            
        for subplot in self._subplot_widget_container.values():
            if subplot.axisSlice() == axis_direction:
                if axis_direction == 2:
                    axis_item = subplot.getAxis("left")
                else:
                    axis_item = subplot.getAxis("bottom")
                axis_item.setLabel(axis_label)
       # self.centralWidget().vb.
      #  self.centralWidget().vb.enableAutoRange(enable=True)
    
    

    def switch_axis_units(self, scale, axis):
        scale = float(scale)
        if isinstance(axis, str):
            if axis == "x":
                ax=0
            elif axis == "y":
                ax=1
            elif axis == "z":
                ax=2
        for axis_item in self.centralWidget().axes[ax]:
            axis_item.setScale(scale)
        
        for subplot in self._subplot_widget_container.values():
            if ax == 2:
                axis_item = subplot.getAxis("left")
                axis_item.setScale(scale)
                axis_item.update()
            else:
                if subplot.axisSlice() == ax:
                    axis_item = subplot.getAxis("bottom")
                    axis_item.setScale(scale)
                    axis_item.update()       
    
    def dataRange(self):
        x_values, y_values, z_values = [], [], []
        for px_type in self.displayed_price_types:
            volatility_data = self.volatility_data_container[px_type]
            for plot_type in self._displayed_assets:
                dataclass = volatility_data.get_dataclass(plot_type)
                x, y, z = dataclass.data()
                x_values = np.append(x_values, x)
                y_values = np.append(y_values, y)
                z_values = np.append(z_values, z.flatten())
        
        x_range = np.nanmin(x_values), np.nanmax(x_values)
        y_range = np.nanmin(y_values), np.nanmax(y_values)
        z_range = np.nanmin(z_values), np.nanmax(z_values)
        return x_range, y_range, z_range
    
    def centralWidget(self) -> PyVolGL3DViewWidget:
        return super().centralWidget()
    
    def showEvent(self, event):
        super().showEvent(event)
        if self._first_show:
            dockwidget = self.findChild(QtWidgets.QDockWidget)
            QtCore.QTimer.singleShot(100, lambda: dockwidget.size().setWidth(self.size().width()/3))
            self._first_show = False
        return 


import sys
from PySide6.QtWidgets import QApplication, QMainWindow
from PySide6 import QtWidgets
from pyvol_terminal.settings.widgets import WindowSettings, SurfaceSettings
from PySide6.QtCore import Slot



class SettingsManager:
    def __init__(self, widget_main=None, widget_layout_v_main=None):
        self.widget_main=widget_main
        self.settings_main = WindowSettings(self)
        self.settings_surface = SurfaceSettings(widget_main=self.widget_main)
        
        
        widget_layout_v_main.addWidget(self.settings_main)
        widget_layout_v_main.addWidget(self.settings_surface)
        
        widget_layout_v_main.setStretchFactor(self.settings_main, 1)
        widget_layout_v_main.setStretchFactor(self.settings_surface, 1)

    @Slot(QtWidgets.QPushButton)
    def on_button_clicked(self, button):
        btn_id = self.settings_main.button_group.id(button)
        self.widget_main.switch_view(self.settings_main.button_name_maps[btn_id])
        self._reset_hide_viewable_widgets()
        match btn_id:
            case 0:
                self.omon_clicked()
            case 1:
                self.vol_table_clicked()
            case 2:
                self.suface_clicked()
            case 3:
                self.smirk_clicked()
            case 4:
                self.term_clicked()
    
    def omon_clicked(self):
        self.splitter_omon.show()
        self.prev_splitter=self.splitter_omon
        
    def vol_table_clicked(self):
        self.splitter_vol_table.show()
        self.prev_splitter=self.splitter_vol_table
                
    def suface_clicked(self):
        
        self.splitter_surface.show()
        
        self.prev_splitter=self.splitter_surface
        
    def smirk_clicked(self):
        pass

    def term_clicked(self):
        pass
    
    def _reset_hide_viewable_widgets(self):
        self.prev_splitter.hide()


class MainWindow(QMainWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setWindowTitle('Volatility Surface')
        self.widget_central = QtWidgets.QWidget()
        self.setCentralWidget(self.widget_central)
        
        self.widget_layout_v_main = QtWidgets.QVBoxLayout(self.widget_central)
        self.settings_surface = SettingsManager(self, self.widget_layout_v_main)
        print(self.widget_layout_v_main.stretch(0))
        print(self.widget_layout_v_main.stretch(1))
        print(self.widget_layout_v_main.stretch(2))

        
    def toggle_subplots(self):
        pass
    def toggle_crosshairs(self):
        pass
    
    def toggle_price_type(self):
        pass
    def toggle_3D_objects(self):
        pass

if __name__ == "__main__":
    app = QApplication(sys.argv)
    mainWin = MainWindow()
    mainWin.show()
    sys.exit(app.exec())
    



import numpy as np
import pyqtgraph as pg
import pyqtgraph.opengl as gl
from PySide6 import QtGui


app = pg.mkQApp()
view = gl.GLViewWidget()
view.show()

# Create 5 random lines (10 points)
pos = np.random.rand(10, 3) * 10  # Example data

# Build transformation matrix (translate + scale)
matrix = QtGui.QMatrix4x4()
matrix.translate(2, 2, 2)  # Shift by (2, 2, 2)
matrix.scale(3, 3, 3)      # Scale by (3, 3, 3)

for i in range(0, len(pos), 2):
    point = QtGui.QVector3D(*pos[i])
    transformed_point = matrix * point
    pos[i] = [transformed_point.x(), transformed_point.y(), transformed_point.z()]

line_plot = gl.GLLinePlotItem(pos=pos, mode="lines", width=2, color='r')
view.addItem(line_plot)

matrix = QtGui.QMatrix4x4()
matrix.translate(2, 2, 2)  # Shift by (2, 2, 2)
matrix.scale(3, 3, 3)      # Scale by (3, 3, 3)



line_plot.setTransform()




import sys
import numpy as np
from PySide6 import QtWidgets, QtCore, QtGui
import pyqtgraph.opengl as gl
from OpenGL import GL

class ContainerItem(gl.GLGraphicsItem.GLGraphicsItem):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        


class GLItemGroup(gl.GLGraphicsItem.GLGraphicsItem, QtWidgets.QGraphicsObject):
    def __init__(self):
        super().__init__(self)
        self._GLGraphicsItem__parent = None 
        self._GLGraphicsItem__children = list() 
        self._GLGraphicsItem__view = None 
        
    def setParentItem(self, item):
        if self._GLGraphicsItem__parent is not None:
            self._GLGraphicsItem__parent._GLGraphicsItem__children.remove(self)
        if item is not None:
            item._GLGraphicsItem__children.append(self)

        if self._GLGraphicsItem__view is not None:
            self._GLGraphicsItem__view.removeItem(self)

        self._GLGraphicsItem__parent = item
        self._GLGraphicsItem__view = None
        
    def childItems(self):
        return list(self._GLGraphicsItem__children)

    def boundingRect(self):
        return QtCore.QRectF()
        
    def paint(self, *args):
        pass
    
    def addItem(self, item):
        item.setParentItem(self)
    

def create_spot(point, color):
    return gl.GLScatterPlotItem(
            pos=point,
            color=color, # Yellow
            size=0.2,
            pxMode=False
        )
def create_constant(point, color):
    return Constant(
            pos=point,
            color=color, # Yellow
            size=0.2,
            pxMode=False
        )

class Constant(gl.GLScatterPlotItem):
    def __init__(self,*args,**kwargs):
        super().__init__(*args, **kwargs)
        
    def paint(self):
        print("\npaint")
        parent = self.parentItem()
        
        if parent is not None:
            # 1. Get parent's full transform matrix
            parent_transform = parent.transform()
            print("parent_transform")
            print(parent_transform)
            viewTransform = parent.viewTransform()
            
            print("viewTransform")
            print(viewTransform)
            arr = np.array(viewTransform.data()).reshape(4, 4)
            print("arr")
            print(arr)
            inv_transform = np.linalg.inv(arr)
            
            print("inv_transform")
            print(inv_transform)
            identity = np.eye(4, dtype=np.float32)
            
            print(identity)
            print(np.array(viewTransform.data()).reshape(4, 4).shape, inv_transform.shape, identity.shape)
            corrected_transform = np.array(viewTransform.data()).reshape(4, 4) @ inv_transform @ identity
            
            print("corrected_transform")
            print(corrected_transform)
            flatten = corrected_transform.flatten().tolist()
            self.setTransform(flatten)
        
        # 6. Render with modified transform
        super().paint()
        
        # 7. Restore original transform if modified
        if parent is not None:
            print("setTransform")
            self.setTransform(np.eye(4))
            print("setTransform2")
            
            

class MainWindow(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()

        self.setWindowTitle("Nested GLScatterPlotItem Example")
        self.setGeometry(100, 100, 800, 600)

        # Create a central widget and a layout
        central_widget = QtWidgets.QWidget()
        self.setCentralWidget(central_widget)
        layout = QtWidgets.QVBoxLayout(central_widget)
        

        # Create a GLViewWidget
        self.gl_view = gl.GLViewWidget()
        self.gl_view.setCameraPosition(distance=40)
        self.showMaximized()
        layout.addWidget(self.gl_view)

        #self.parent_container = ContainerItem()
        self.parent_container = GLItemGroup()

        pos = np.random.random(size=(500, 3))
        
        pos *= [10, 10, 10]
        pos[0] = (0, 0, 0)
        
        pos = 0.5*np.array([[1, 1, 1]])

        self.scatter_plot_item1 = create_spot(pos, (1, 1, 0, 1))
        self.scatter_plot_item1.setParentItem(self.parent_container)
        pos = np.array([[1, 1, 1]])
        self.parent_container = create_spot(pos, (1, 0, 0, 1))
        
        
        self.scatter_plot_item1.setParentItem(self.parent_container)

        
        self.gl_view.addItem(self.parent_container)

        
        grid = gl.GLGridItem()
        grid.scale(2, 2, 1)
        self.gl_view.addItem(grid)
        
        pos = np.array([[1,1,1]])
        print(f"parent_container: {self.parent_container.visible()}")
        print(f"scatter_plot_item1: {self.scatter_plot_item1.visible()}")
        
        self.parent_container.setVisible(False)
        
        
        print(f"parent_container: {self.parent_container.visible()}")
        print(f"scatter_plot_item1: {self.scatter_plot_item1.visible()}")

        self.scatter_plot_item1.setVisible(True)
        print(f"parent_container: {self.parent_container.visible()}")
        print(f"scatter_plot_item1: {self.scatter_plot_item1.visible()}")
        import sys 
        sys.exit()

        
        for i in range(1, 11):
            x = gl.GLTextItem(pos = (i/10, 0, 0), text=f"{round(i/10, 1)}", color="white")
            y = gl.GLTextItem(pos = (0, i/10, 0), text=f"{round(i/10, 1)}", color="yellow")
            z = gl.GLTextItem(pos = (i/10, 0, i/10), text=f"{round(i/10, 1)}", color="cyan")
            self.gl_view.addItem(x)
            self.gl_view.addItem(y)
            self.gl_view.addItem(z)
            
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.update_plots)
        self.counter=1

    def update_plots(self):
        self.counter+=0.15
        tr = self.parent_container.transform()
        tr.scale(self.counter, self.counter, self.counter)
        
        print("\n")
        print(f"parent_container: {tr}")
        for child in self.parent_container.childItems():
            
            print(f"child tr: {child.transform()}")

        self.gl_view.update()

            


def main():
    """Main function to run the application."""
    app = QtWidgets.QApplication(sys.argv)
    window = MainWindow()
    window.timer.start(1000)
    sys.exit(app.exec())

if __name__ == "__main__":
    main()


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, ClassVar
if TYPE_CHECKING:
    from .GL3DViewBox import GL3DViewBox
    from ..widgets.GL3DViewWidget import GL3DViewWidget



    
from PySide6 import QtWidgets, QtCore, QtGui
from abc import ABC, abstractmethod, ABCMeta
import abc
from pyqtgraph import Transform3D, functions as fn
from pyqtgraph.opengl.GLGraphicsItem import GLGraphicsItem
from PySide6.QtWidgets import QGraphicsItem
from enum import Enum, EnumType, Flag
import warnings
from pyqtgraph.Qt import QT_LIB
import weakref
from .GL3DGraphicsItemMixin import GL3DGraphicsItemMixin
from dataclasses import dataclass, field, InitVar
import numpy as np
import traceback
import inspect

from pprint import pprint
import traceback
import io


def stackprinter_print_stack(limit, phrase):
    buf = io.StringIO()
    
    # Print stack to buffer (returns None, but writes to buf)
    traceback.print_stack(limit=limit, file=buf)
    
    # Get the value from buffer and split lines
    stack_trace = buf.getvalue()
    lines = stack_trace.split('\n')
    
    # Process each line to remove the phrase
    processed_lines = []
    for line in lines:
        processed_lines.append(line.split(str(phrase))[-1])
    
    # Join the processed lines
    return "\n".join(processed_lines)


    
@dataclass(slots=True)
class _ABCBasePlotDataset(ABC):
    initData: InitVar[np.ndarray|Tuple[np.ndarray, ...]] = None
    allFinite: List[Optional[bool]] = field(default_factory=list)
    _dataBounds: List[List[float]] | None = field(init=False, default=None)
    
    def __post_init__(self, initData):
        self._initialize_attributes()
        
        
    def _initialize_attributes(self):
        if len(self.allFinite) == 0:
            self.allFinite = [None] * self.num_components
            self._updateDataBounds()
        
    def dataBounds(self) -> QtCore.QRectF | None:
        if self._dataBounds is None: 
            self._updateDataBounds()
        return self._dataBounds
    
    @abstractmethod
    def data(self) -> np.ndarray | Tuple[np.ndarray,...]:
        ...
        
    @abstractmethod
    def dataKwargs(self) -> Dict[str, np.ndarray]:
        ...

    @property
    @abstractmethod
    def num_components(self) -> int:
        """Number of data components (e.g., 3 for x/y/z)"""
        
    @abstractmethod
    def getComponents(self) -> List[np.ndarray]:
        """Get list of component arrays (flattened for bounds calculation)"""
    
    def _updateDataBounds(self):
        components = self.getComponents()
        if any([values is None for values in components]):
            return None
         
        dataBounds=[]
        for idx, arr in enumerate(components):
            minRange, maxRange, finiteFlag = self._getArrayBounds(arr, self.allFinite[idx])
            dataBounds.append((minRange, maxRange))
            self.allFinite[idx]=finiteFlag
        self._dataBounds=dataBounds
                
    def _getArrayBounds(self,
                        arr: np.ndarray,
                        all_finite: bool | None
                        ) -> tuple[float, float, bool]:
        # here all_finite could be [None, False, True]
        if not all_finite:  # This may contain NaN or inf values.
            # We are looking for the bounds of the plottable data set. Infinite and Nan
            # are ignored.
            selection = np.isfinite(arr)
            # True if all values are finite, False if there are any non-finites
            all_finite = bool(selection.all())
            if not all_finite:
                arr = arr[selection]
        
        # here all_finite could be [False, True]
        try:
            amin = np.min( arr )  # find minimum of all finite values
            amax = np.max( arr )  # find maximum of all finite values
        except ValueError:  # is raised when there are no finite values
            amin = np.nan
            amax = np.nan
        return amin, amax, all_finite

@dataclass(slots=True)
class PlotDatasetFlatMesh(_ABCBasePlotDataset):    
    _pos: np.ndarray = field(init=False, default=None)
    
    def __post_init__(self, initData):
        self._pos = initData
        super(PlotDatasetFlatMesh, self).__post_init__(initData)
        
    @property
    def num_components(self) -> int:
        return 3
        
    def getComponents(self) -> List[np.ndarray]:
        
        return [self._pos[:, i] for i in range(self.num_components)]
    
    @property
    def pos(self) -> np.ndarray:
        return self._pos
        
    @pos.setter
    def pos(self, value: np.ndarray):
        self._pos = value

    def data(self) -> np.ndarray:
        return self.pos
    
    def dataKwargs(self):
        return {"pos" : self.pos}
            
@dataclass(slots=True)
class PlotDatasetMesh(_ABCBasePlotDataset):
    initData: InitVar[Tuple[np.ndarray, ...]] = None
    
    _x: np.ndarray = field(init=False, default=None)
    _y: np.ndarray = field(init=False, default=None)
    _z: np.ndarray = field(init=False, default=None)
    
    def __post_init__(self, initData):
        self.x, self.y, self.z = initData
        super(PlotDatasetMesh, self).__post_init__(initData)
            
    @property
    def num_components(self) -> int:
        return 3
        
    def getComponents(self) -> List[np.ndarray]:
        return [c.ravel() for c in (self.x, self.y, self.z)]
    
    @property
    def x(self) -> np.ndarray:
        return self._x
        
    @x.setter
    def x(self, value: np.ndarray):
        self._x = value
    
    @property
    def y(self) -> np.ndarray:
        return self._y
        
    @y.setter
    def y(self, value: np.ndarray):
        self._y = value
    
    @property
    def z(self) -> np.ndarray:
        return self._z
        
    @z.setter
    def z(self, value: np.ndarray):
        self._z = value

    def data(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        return (self.x, self.y, self.z)

    def dataKwargs(self):
        return {"x" : self.x, "y" : self.y, "z" : self.z}



@dataclass(slots=True)
class CacheData:
    _null3: ClassVar[List[None]] = [None, None, None]
    _false3: ClassVar[List[None]] = [False, False, False]
    _null32: ClassVar[List[List[None]]] = [[None, None],[None, None], [None, None]]
    
    _mask: List[None]|List[np.ndarray] = field(init=False, default_factory=lambda : CacheData._null3.copy())
    _bounds: List[List[None]]|List[List[float]] = field(init=False, default_factory=lambda : [row.copy() for row in CacheData._null32])
    _mask: List[None]|List[np.ndarray] = field(init=False, default_factory=lambda : CacheData._null3.copy())
    _orthoRange: List[None]|List[float] = field(init=False, default_factory=lambda : CacheData._null3.copy())
    _frac: List[None]|List[float] = field(init=False, default_factory=lambda : CacheData._null3.copy())
    _cacheFlag: List[bool] = field(init=False, default_factory=lambda : CacheData._false3.copy())

    def __repr__(self):
        out = {}
        for name in self.__slots__:
            if name.startswith('_'):
                public_name = name[1:]
                method = getattr(self, public_name, None)
                if callable(method):
                    try:
                        out[public_name] = method()
                    except Exception:
                        out[public_name] = "<error>"
        return f"{self.__class__.__name__}({out})"   
     
    def cacheFlag(self, ax=None):
        if ax is None:
            return self._cacheFlag
        else:
            return self._cacheFlag[ax]
        
    
    def setCacheFlag(self, flag, ax=None):
        if ax is None:
            self._cacheFlag=flag
        else:
            self._cacheFlag[flag] = ax

    def setBounds(self, bounds, ax=None):
        if ax is None:
            self._bounds = bounds
            self.setCacheFlag([not None in lim for lim in bounds])
        else:
            self._bounds[ax] = bounds
            self.setCacheFlag(bounds, ax)

    def bounds(self, ax=None):
        if ax is None:
            return self._bounds
        else:
            return self._bounds[ax]

    def setMask(self, mask, ax=None):
        if ax is None:
            self._mask=mask     
        else:
            self._mask[ax] = mask
    
    def mask(self, ax=None):
        if ax is None:
            self._mask
        else:
            return self._mask[ax]

    def setOrthoRange(self, orthoRange, ax=None):
        if ax is None:
            self._orthoRange=orthoRange
        else:
            self._orthoRange[ax] = orthoRange

    def orthoRange(self, ax=None):
        if ax is None:
            return self._orthoRange
        else:
            return self._orthoRange[ax] 

    def setFrac(self, frac, ax=None):
        if ax is None:
            self._frac=frac
        else:
            self._frac[ax] = frac
        
    def frac(self, ax=None):
        if ax is None:
            return self._frac
        else:
            return self._frac[ax] 

    def reset(self):
        self.setBounds([row.copy() for row in self._null32])
        self.setCacheFlag(self._false3.copy())
        self.setMask(self._null3.copy())
        self.setOrthoRange(self._null3.copy())
        self.setFrac(self._null3.copy())



class MeshPlotitemProperties:
    def __init__(self, *args, **kwargs):
        self._x: np.ndarray
        self._y: np.ndarray 
        self._z: np.ndarray
        super().__init__(*args, **kwargs)
    
    @property
    def x(self):
        return self._x
      
    @x.setter
    def _x(self, value):
        self._x = value
        
    @property
    def y(self):
        return self._y
      
    @y.setter
    def _y(self, value):
        self._y = value

    @property
    def z(self):
        return self._z
      
    @z.setter
    def _z(self, value):
        self._z = value

class VectorPlotitemProperties:
    def __init__(self, *args, **kwargs):
        self._pos: np.ndarray
        super().__init__(*args, **kwargs)
    
    @property
    def pos(self):
        
        return self._pos
      
    @pos.setter
    def _pos(self, value):
        if (self.pos != value).any():
            
            self._pos = value


class BaseGL3DPlotDataItemMixin(GL3DGraphicsItemMixin):
    sigPlotChanged = QtCore.Signal(object) 
    
    def __init__(self, PlotDatasetClass, parentItem=None, **kwargs):
        self.last=None
        self._dataset: _ABCBasePlotDataset=None
        self._datasetDisplay: _ABCBasePlotDataset=None
        self._cacheData = CacheData()
        self._cacheValid=False
        self._viewBox: GL3DViewBox=None
        _connected_points = self._determineShading()
        
        if _connected_points:
            zFilterMode = "clip"
        else:
            zFilterMode = "filter"
        
        self.data_opts = {"zFilterMode" : zFilterMode,
                          "connectedPoints" : _connected_points,
                          "PlotDatasetClass" : PlotDatasetClass,
                          }
        super().__init__(parentItem=parentItem, **kwargs)
        

        self.setProperty('xViewRangeWasChanged', False)
        self.setProperty('yViewRangeWasChanged', False)
        self.setProperty('zViewRangeWasChanged', False)
        self.setProperty('styleWasChanged', True)
   #     self.setParentItem(parentItem)
    

    def _determineShading(self):
        if hasattr(self, 'smooth') and self.smooth:
            return True
        if hasattr(self, 'computeNormals') and self.computeNormals:
            return True
        if hasattr(self, 'mode'):
            return self.mode in ['triangles', 'triangle_strip']
        return False

    def viewTransform2(self):
        view = self.getViewBox()
        if view is None:
            return None
        if hasattr(view, 'implements') and view.implements('ViewBox'):
            return self.itemTransform(view.innerSceneItem())[0]
        else:
            return self.sceneTransform()

    def setViewBox(self, vb):
        self._viewBox=vb

    def dataRect(self): 
        return None if self._dataset is None else self._dataset.dataBounds()
    
    def viewTransformChanged(self):
        self.invalidateBounds()
        #self.prepareGeometryChange()

    def invalidateBounds(self):
        self._cacheData.reset()
        
    @QtCore.Slot(object, object)
    @QtCore.Slot(object, object, object)
    def viewRangeChanged(self, vb=None, ranges=None, changed=None):
        update_needed = False
        if changed is None or changed[0]: 
            self.setProperty('xViewRangeWasChanged', True)
            self._datasetDisplay = None
            update_needed = True
        if changed is None or changed[1]:
            self.setProperty('yViewRangeWasChanged', True)
            self._datasetDisplay = None
            update_needed = True
        if changed is None or changed[2]:
            self.setProperty('zViewRangeWasChanged', True)
            self._datasetDisplay = None
            update_needed = True

        if update_needed:
            self.updateItems(styleUpdate=False)
            
    def updateItems(self, styleUpdate: bool = True):
        dataset = self._getDisplayDataset()
        if dataset is None:
            self.hide()
        else:
            self.updateData()
    
    def updateData(self, **kwargs):
        self.invalidateBounds()
        #self.prepareGeometryChange()
        self.informViewBoundsChanged()
        self.sigPlotChanged.emit(self)
    
    def getData(self):
        dataset = self._dataset
        view = self.view()
        if dataset is None or view is None:
            return (None, None, None)
        
 
        autoRange = view.vb.state['autoRange']
        if all(autoRange):
            return dataset.data()
        else:
            datasetDisplayed = self._getDisplayDataset()
            if not datasetDisplayed is None:
                return datasetDisplayed.data()
            return (None, None, None)
    


            
    def getOriginalDataset(self) -> tuple[None, None] | tuple[np.ndarray, np.ndarray, np.ndarray]:
        dataset = self._dataset
        return (None, None, None) if dataset is None else dataset.data()
    
    def _getDisplayDataset(self) -> _ABCBasePlotDataset | None:
        if self._dataset is None:
            self.last=1
            self._cacheValid=False
            return None
        # Return cached processed dataset if available and still valid:
        if all((not self._datasetDisplay is None,
                not self.property('xViewRangeWasChanged'),
                not self.property('yViewRangeWasChanged'),
                not self.property('zViewRangeWasChanged')
                )):
            self._cacheValid=True
            self.last=2
                
            return self._datasetDisplay
        self._cacheValid=False
        allFinite = self._dataset.allFinite
        data = self._dataset.data()        
        view = self.getViewBox()

        if view is None:
            view_range = None
        else:
            view_range = view.viewRange()  # this is always up-to-date
        if view_range is None:
            view_range = self.viewRect()
 
        if view is None or view.autoRangeEnabled()[0]:
            pass  # no ViewBox to clip to, or view will autoscale to data range.
                
        if not all(allFinite):
            data = self._dataset.data()
            self.last=3
            
        if not view_range is None and all(allFinite):
            if not any(view.autoRangeEnabled()):
                data = self.clipFromView(data, view_range)        
                self.last=4
                    
        self._datasetDisplay = self.data_opts["PlotDatasetClass"](data, allFinite)
        
        self.setProperty('xViewRangeWasChanged', False)
        self.setProperty('yViewRangeWasChanged', False)
        self.setProperty('zViewRangeWasChanged', False)
        return self._datasetDisplay

    def dataBounds(self,
                    frac=(1., 1. ,1),
                    orthoRange=(None, None, None)
                    ) -> Tuple[float, float, float]:

        if not self.visible() or self._dataset is None:
            return [[None, None], [None, None], [None, None]]
        else:
            data = self._dataset.data()
            mask = np.ones(len(data), dtype=bool)

            cache_bounds = {}
            mask_container={}

            for ax in range(2):
                if self._cacheData.cacheFlag(ax) is None and all((self._cacheData.frac(ax) == frac[ax],
                                                                  self._cacheData.orthoRange(ax) == orthoRange[ax])):
                    cache_bounds[ax] = self._cacheData.bounds(ax)
                    mask_container[ax] = self._cacheData.mask(ax)

                if not orthoRange[ax] is None:
                    mask = (data[:, ax] >= orthoRange[ax][0]) & (data[:, ax] <= orthoRange[ax][1])
                    self._cacheData.setMask(mask, ax)
                    mask_container[ax] = mask
                else:
                    self._cacheData.setMask(None, ax)
            
            x_filtered = data[0][mask_container[0]] if 0 in mask_container else data[0]
            y_filtered = data[1][mask_container[1]] if 1 in mask_container else data[1]
        
            if self._cacheData.cacheFlag(2) is None and all((self._cacheData.frac(2) == frac[2],
                                                             self._cacheData.orthoRange(2) == orthoRange[2]
                                                             )
                                                            ):
                cache_bounds[2] = self._cacheData.bounds(2)
                
            z_filtered = data[2]
            if 0 in mask_container:
                z_filtered = z_filtered[mask_container[0], :]
            
            if 1 in mask_container:
                z_filtered = z_filtered[:, mask_container[1]]

            if not orthoRange[2] is None:
                z_filtered = np.clip(z_filtered, orthoRange[2][0], orthoRange[2][1])

            limits = [[bound.min(), bound.max()] if i not in cache_bounds else cache_bounds[i] for i, bound in enumerate((x_filtered, y_filtered, z_filtered))]
            self._cacheData.setBounds(limits)
            return self._cacheData.bounds()
        
        
    def setData(self, *args, **kwargs):
        sig = inspect.signature(self._setDataChild)

        bound = sig.bind(*args, **kwargs)
        bound.apply_defaults()  

        call_args = []
        call_kwargs = {}

        for name, param in sig.parameters.items():
            if param.kind in (
                inspect.Parameter.POSITIONAL_ONLY,
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
            ):
                call_args.append(bound.arguments[name])
            elif param.kind == inspect.Parameter.VAR_POSITIONAL:
                call_args.extend(bound.arguments.get(name, ()))
            elif param.kind == inspect.Parameter.KEYWORD_ONLY:
                call_kwargs[name] = bound.arguments[name]
            elif param.kind == inspect.Parameter.VAR_KEYWORD:
                call_kwargs.update(bound.arguments.get(name, {}))

        self.blockUpdates(True)
        self._setDataChild(*call_args, **call_kwargs)
        if isinstance(self.dataAttr, (list, tuple)):
            if all([not attr is None for attr in self.dataAttr]):
                self._dataset = self.data_opts["PlotDatasetClass"](self.dataAttr)
        else:
            if not self.dataAttr is None:    
                self._dataset = self.data_opts["PlotDatasetClass"](self.dataAttr)
        self._datasetDisplay = None
        self.updateItems(styleUpdate=self.property('styleWasChanged'))
        self.blockUpdates(False)
        self.update()
        
    @abstractmethod
    def clipFromView(self, dataset, viewrange):...
    
    @abstractmethod
    def _setDataChild(self, *args, **kwargs):...

    @property
    @abstractmethod
    def dataAttr(self):...


class GL3DMeshPlotDataItemMixin(BaseGL3DPlotDataItemMixin):
    topology_type="mesh"
    def __init__(self, parentItem=None, **kwargs):
        super().__init__(PlotDatasetClass=PlotDatasetMesh,
                        parentItem=parentItem,
                        **kwargs
                        )
        
    def _paintHelper(self):
        x, y, z = self.getData()
        
        if np.array_equal(self._x, x):
           x=None 
        if np.array_equal(self._y, y):
           y=None 
        if np.array_equal(self._z, z):
           z=None 

        self.blockUpdates(True)

        self._setDataChild(x=x, y=y, z=z)
        self.blockUpdates(False)
        
    def clipFromView(self, dataset, view_range):
        x, y, z = dataset
        
        x_min, x_max = view_range[0]
        y_min, y_max = view_range[1]
        z_min, z_max = view_range[2]
        
        x_mask = (x >= x_min) & (x <= x_max)
        y_mask = (y >= y_min) & (y <= y_max)
        
        x_filtered = x[x_mask]
        y_filtered = y[y_mask]
                
        z_filtered = z[x_mask, :][:, y_mask]
        z_filtered = np.clip(z_filtered, z_min, z_max)
        return x_filtered, y_filtered, z_filtered

class GL3DFlatMeshPlotDataItemMixin(BaseGL3DPlotDataItemMixin):
    topology_type="FlatMesh"
    
    def __init__(self, parentItem=None, **kwargs):
        super().__init__(PlotDatasetClass=PlotDatasetFlatMesh,
                         parentItem=parentItem,
                         **kwargs
                         )
        
    def _paintHelper(self):
        pos = self.getData()
        
        if np.array_equal(self.pos, pos):
           pos=pos 

        self.blockUpdates(True)
        self._setDataChild(pos=pos)
        self.blockUpdates(False)

    def clipFromView(self, dataset, view_range):
        if self.data_opts["zFilterMode"]=="clip":
            mask = ((dataset[:, 0] >= view_range[0][0]) & (dataset[:, 0] <= view_range[0][1])
                   &(dataset[:, 1] >= view_range[1][0]) & (dataset[:, 1] <= view_range[1][1]) 
                   )
            z_filtered = dataset[mask]
            z_filtered = np.clip(z_filtered, view_range[2][0], view_range[2][1])
        else:
            mask = ((dataset[:, 0] >= view_range[0][0]) & (dataset[:, 0] <= view_range[0][1])
                   &(dataset[:, 1] >= view_range[1][0]) & (dataset[:, 1] <= view_range[1][1]) 
                   &(dataset[:, 2] >= view_range[2][0]) & (dataset[:, 2] <= view_range[2][1])
                   )
            z_filtered = dataset[mask]
        return z_filtered

    def dataBounds(self,
                    frac=(1., 1. ,1),
                    orthoRange=(None, None, None)
                    ) -> Tuple[float, float, float]:
        
        if not self.visible() or self._dataset is None:
            return [[None, None], [None, None], [None, None]]
        else:
            data = self._dataset.data()
            data_t = data[:, 0], data[:, 1], data[:, 2]
            mask = np.ones(len(data), dtype=bool)
            cache_bounds = {}
            mask_container={}

            for ax in range(3):
                if self._cacheData.cacheFlag(ax) is None and all((self._cacheData.frac(ax) == frac[ax],
                                                                  self._cacheData.orthoRange(ax) == orthoRange[ax])):
                    cache_bounds[ax] = self._cacheData.bounds(ax)
                    mask_container[ax] = self._cacheData.mask(ax)
            
                if not orthoRange[ax] is None:
                    mask = (data_t[ax] >= orthoRange[ax][0]) & (data_t[ax] <= orthoRange[ax][1])
                    self._cacheData.setMask(mask, ax)
                    mask_container[ax] = mask
                else:
                    self._cacheData.setMask(None, ax)
            
            x_filtered = data_t[0][mask_container[0]] if 0 in mask_container else data_t[0]
            y_filtered = data_t[1][mask_container[1]] if 1 in mask_container else data_t[1]
            z_filtered = data_t[2][mask_container[2]] if 2 in mask_container else data_t[2]
            limits = [[bound.min(), bound.max()] if i not in cache_bounds else cache_bounds[i] for i, bound in enumerate((x_filtered, y_filtered, z_filtered))]
            self._cacheData.setBounds(limits)
            return self._cacheData.bounds()
        



from __future__ import annotations 
from typing import List, Optional, Any, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    
    from pandas import DataFrame
    from ..utils import ws_pyvol_keymap


from .utils import InstrumentManager, InstrumentContainer, BaseMap, OptionMap, ManagerMap
import pandas as pd
from collections import defaultdict
from .instruments import ABCInstrument, Option, Future, Spot, SpotSpecifications, FutureSpecifications, OptionInvSpecifications


def create_instrument_manager(instruments_config,
                              name=None,
                              ws_pyvol_map: ws_pyvol_keymap=None,
                              ) -> InstrumentManager:
    objects_container_all={}
    name_type_container_all = {}
    instrument_manager_kwargs = {"config" : instruments_config}
    
    #create_option_pair_objects(instruments_config["option"]["data"])
    
    for instrument_type in instruments_config:
        if instrument_type in instruments_config:
            
            df: pd.DataFrame = instruments_config[instrument_type]["data"]
            df = df.rename(columns={ws_pyvol_map.ticker : "ticker"})
            
            
            instrument_container, name_type_container = globals()[f"create_{instrument_type}_objects"](df,
                                                                                                       instruments_config,
                                                                                                       objects_container_all)

            objects_container_all = objects_container_all | instrument_container.objects
            name_type_container_all = name_type_container_all | name_type_container
            instrument_manager_kwargs[f"{instrument_type}_instrument_container"] = instrument_container
    manager_map = ManagerMap(name_type_container_all)
    instrument_manager_kwargs["manager_map"]=manager_map
    instrument_manager_kwargs["all_instrument_objects"]=objects_container_all
    instrument_manager_kwargs["name"]=name
    return InstrumentManager(**instrument_manager_kwargs)

def _create_maps(instrument_object_dict):        
    index_name_map = {idx: name for idx, name in enumerate(instrument_object_dict)}
    name_index_map = {name: idx for idx, name in enumerate(instrument_object_dict)}
    return BaseMap(index_name_map, name_index_map)

def create_spot_objects(df, config, *args):
    object_container = {}
    ticker_category_map={}
    if not df is None:
        ticker = df["ticker"].item()        
        kwargs = {key : value for key, value in config["spot"].items()}
        specs = SpotSpecifications(ticker=ticker,
                                   px_quotation="USD"
                                   )
        cls = config["spot"]["object"]
        instrument_object = cls(specs, **kwargs)
        ticker_category_map[ticker] = "spot"
        object_container[ticker] = instrument_object
            
    maps = _create_maps(object_container)
    spot_container = InstrumentContainer(object_container, config["spot"]["price_types"], maps, "spot", df)
    return spot_container, ticker_category_map
    

def create_future_objects(df, config, other_objects: Dict[str, ABCInstrument]):   
    ticker_category_map = {}
    object_container = {}
    for future_row in df.iterrows():
        future_row = future_row[1]
        ticker = future_row["ticker"]
        expiry = future_row["expiry"]
        ticker_category_map[ticker] = "future"    

        if "underlying_map" in config["future"]:
            if isinstance(config["future"]["underlying_map"], str):
                underlying_ticker=config["future"]["underlying_map"][ticker]
                underlying_object = other_objects[underlying_ticker]
            elif isinstance(config["future"]["underlying_map"], Dict):
                underlying_ticker=config["future"]["underlying_map"][ticker]
                underlying_object = other_objects[underlying_ticker]
            else:
                raise KeyError("You have specified an underyling map, but the underlying object has not been instantiated yet. Make sure underlying objects are instantiated first.")
        else:   
            underlying_object=None
      #  args = [ticker, expiry]
        kwargs = config["future"]
        kwargs["ticker"] = ticker
        kwargs["expiry"] = expiry
        kwargs["underlying_ticker"] = underlying_ticker
        kwargs["specs"] = FutureSpecifications(ticker=ticker,
                                               px_quotation="USD"
                                               )

        if "interest_rate_config" in config["future"]:
            kwargs["interest_rate_engine"] = config["future"]["interest_rate_config"]["engine"]
        if "dividend_rate_config" in config["future"]:
            kwargs["dividend_rate_engine"] = config["future"]["dividend_rate_config"]["engine"]
        cls = config["future"]["object"]
        instrument_object = cls(**kwargs)        
        object_container[ticker]=instrument_object
            
    maps = _create_maps(object_container)
    futures_instrument_container = InstrumentContainer(object_container, config["future"]["price_types"], maps, "future", df)
    return futures_instrument_container, ticker_category_map


def create_option_objects(df, config, other_objects: Dict[str, ABCInstrument|List[ABCInstrument]]):
    object_container = {}
    ticker_category_map = {}
    
    df["strike"] = pd.to_numeric(df["strike"], errors='coerce')
    df["expiry"] = pd.to_numeric(df["expiry"], errors='coerce')
    underlying_px_type = config["option"]["underyling_price_type"]
    
    paired_options, pairs = create_option_pair_objects(df)
    
    del config["option"]["underyling_price_type"]
    print(other_objects.keys())
    objects_to_be_paired = {}
    name_underlying_object = {}
    for _, option_row in df.iterrows():
        ticker = option_row["ticker"]
        
        if "underlying_map" in config["option"]:
            if isinstance(config["option"]["underlying_map"], str):
                underlying_object = other_objects[config["option"]["underlying_map"][ticker]]
            elif isinstance(config["option"]["underlying_map"], Dict):
                underlying_object = other_objects[config["option"]["underlying_map"][ticker]]
        else:
            underlying_object_list: List[ABCInstrument]=[]
            for ob in other_objects:
                if ob.instrument_type=="spot":
                    underlying_object_list.append(ob)
            if len(underlying_object_list) != 1:
                raise KeyError("You need to specify an underyling for the options if you have multiple spot objects")
            else:
                underlying_object = underlying_object_list[0]
                
        ticker_category_map[ticker] = "option"    
        
        flag=option_row["flag"]
        
        args = [option_row["strike"], option_row["expiry"], flag]
        kwargs = {"interest_rate_engine" : config["option"]["interest_rate_config"]["engine"],
                  "dividend_rate_engine" : config["option"]["dividend_rate_config"]["engine"]}

        opt_engine = config["option"]["engine"](*args, **kwargs)
        name_underlying_object[ticker] = underlying_object
        
        specs = O
        
        cls = config["option"]["object"]
        
        instrument_object = cls(ticker=ticker,
                                underlying_ticker=underlying_object.ticker,
                                underlying_px_type=underlying_px_type,
                                strike=option_row["strike"],
                                expiry=option_row["expiry"],
                                flag=option_row["flag"],
                                option_engine=opt_engine,
                                **config["option"]
                                )
        underlying_object.add_child_derivative_callback(instrument_object.update_from_underlying)
        object_container[ticker]=instrument_object
        
    
    options_map = _create_option_map(df, object_container)
    
   # object_container.update(pair_objects)
    
                    
    options_instrument_container = InstrumentContainer(object_container,
                                                       config["option"]["price_types"],
                                                       options_map,
                                                       "option",
                                                       df)   
    return options_instrument_container, ticker_category_map

    
def create_option_pair_objects(df: DataFrame):
    grouped = df.groupby(['strike', 'expiry'])
    pairs = []
    for _, group in grouped:
        if set(group['flag']) >= {'c', 'p'}:
            # Create all c-p pairs
            calls = group[group['flag'] == 'c']
            puts = group[group['flag'] == 'p']
            for _, call_row in calls.iterrows():
                for _, put_row in puts.iterrows():
                    pairs.append([call_row['ticker'], put_row['ticker']])

    has_both_flags = grouped['flag'].transform(lambda x: set(x) >= {'c', 'p'})

    df_filtered = df[has_both_flags]
    
    
    return df_filtered["ticker"].to_list(), pairs

def _create_option_map(df: DataFrame,
                       object_container: Dict[str, Option],
                       ) -> OptionMap:
    grouped = df.groupby(['strike', 'expiry'])
    put_call_pair_map = {}
    for (strike, expiry), group in grouped:
        calls = group[group['flag'] == 'c']
        puts = group[group['flag'] == 'p']
        
        for _, row in calls.iterrows():
            name_call = row['ticker']
            name_put = puts['ticker'].values[0] if not puts.empty else None
            put_call_pair_map[name_call] = name_put
            if not name_put is None:
                opt_put = object_container[name_put]
                opt_call = object_container[name_call]
                opt_put.add_metric_callbacks(opt_call.update_from_OTM_sister)
                opt_call.add_metric_callbacks(opt_put.update_from_OTM_sister)

        for _, row in puts.iterrows():
            name_put = row['ticker']
            name_call = calls['ticker'].values[0] if not calls.empty else None
            put_call_pair_map[name_put] = name_call
            
    underlying_ticker_map={}
    name_underlying_map={}

    for name, opt in object_container.items():
        underlying_ticker = opt.underlying_ticker
        if not name in name_underlying_map:
            name_underlying_map[name]=underlying_ticker
        if not underlying_ticker in underlying_ticker_map:
            underlying_ticker_map[underlying_ticker] = [name]
        else:
            underlying_ticker_map[underlying_ticker].append(name)
            
    expiry_strike_map = defaultdict(set)
    strike_expiry_map = defaultdict(set)
    expiry_instrument_map = defaultdict(list)
    strike_instrument_map = defaultdict(list)
    expiry_strike_instrument_map = defaultdict(lambda: defaultdict(list))
    strike_expiry_instrument_map = defaultdict(lambda: defaultdict(list))
    expiry_strike_type_instrument_map = defaultdict(lambda: defaultdict(dict))
    strike_expiry_type_instrument_map = defaultdict(lambda: defaultdict(dict))
    type_expiry_strike_instrument_map = defaultdict(lambda: defaultdict(dict))

    for _, row in df.iterrows():
        expiry = row["expiry"]
        strike = row["strike"]
        ticker = row["ticker"]
        flag = row["flag"]
        
        expiry_strike_map[expiry].add(strike)
        strike_expiry_map[strike].add(expiry)
        expiry_instrument_map[expiry].append(ticker)
        strike_instrument_map[strike].append(ticker)
        expiry_strike_instrument_map[expiry][strike].append(ticker)
        strike_expiry_instrument_map[strike][expiry].append(ticker)

        if ticker in put_call_pair_map:
            pair_instr = put_call_pair_map[ticker]
            if flag == 'c':
                expiry_strike_type_instrument_map[expiry][strike]["c"] = ticker
                expiry_strike_type_instrument_map[expiry][strike]["p"] = pair_instr
                strike_expiry_type_instrument_map[strike][expiry]["c"] = ticker
                strike_expiry_type_instrument_map[strike][expiry]["p"] = pair_instr
                type_expiry_strike_instrument_map["c"][expiry][strike] = ticker
                type_expiry_strike_instrument_map["p"][expiry][strike] = pair_instr
            elif flag == 'p':
                expiry_strike_type_instrument_map[expiry][strike]["p"] = ticker
                expiry_strike_type_instrument_map[expiry][strike]["c"] = pair_instr
                strike_expiry_type_instrument_map[strike][expiry]["p"] = ticker
                strike_expiry_type_instrument_map[strike][expiry]["c"] = pair_instr
                type_expiry_strike_instrument_map["p"][expiry][strike] = ticker
                type_expiry_strike_instrument_map["c"][expiry][strike] = pair_instr

    map_kwargs={}
    
    map_kwargs["index_name_map"] = {idx: name for idx, name in enumerate(object_container)}
    map_kwargs["name_index_map"] = {name: idx for idx, name in enumerate(object_container)}
    map_kwargs["underlying_ticker_map"]=underlying_ticker_map
    map_kwargs["name_underlying_map"]=name_underlying_map

    map_kwargs["expiry_strike_map"] = {k: sorted(v) for k, v in expiry_strike_map.items()}
    map_kwargs["strike_expiry_map"] = {k: sorted(v) for k, v in strike_expiry_map.items()}
    map_kwargs["expiry_instrument_map"] = dict(expiry_instrument_map)
    map_kwargs["strike_instrument_map"] = dict(strike_instrument_map)
    map_kwargs["expiry_strike_instrument_map"] = {k: {sk: sv for sk, sv in v.items()} for k, v in expiry_strike_instrument_map.items()}
    map_kwargs["strike_expiry_instrument_map"] = {k: {sk: sv for sk, sv in v.items()} for k, v in strike_expiry_instrument_map.items()}
    map_kwargs["expiry_strike_type_instrument_map"] = {k: {sk: dict(sv) for sk, sv in v.items()} for k, v in expiry_strike_type_instrument_map.items()}
    map_kwargs["strike_expiry_type_instrument_map"] = {k: {sk: dict(sv) for sk, sv in v.items()} for k, v in strike_expiry_type_instrument_map.items()}
    map_kwargs["type_expiry_strike_instrument_map"] = {k: {sk: dict(sv) for sk, sv in v.items()} for k, v in type_expiry_strike_instrument_map.items()}
    map_kwargs["put_call_map"] = put_call_pair_map
    return OptionMap(**map_kwargs)





from PySide6 import QtWidgets, QtCore, QtGui

import sys
import weakref
import numpy as np
from OpenGL.GL import * 
import pyqtgraph as pg
from pyqtgraph import opengl
import warnings
from pyqtgraph.Qt import QT_LIB, QtCore, QtWidgets
from pyvol_terminal.gl_3D_graphing import GLGraphicsItem

from pyvol_terminal.gl_3D_graphing.widgets.GLScatterPlotItem import GLScatterPlotItem 




def get_item_data(item):
    if isinstance(item, opengl.GLScatterPlotItem):
        return item.pos
    elif isinstance(item, opengl.GLMeshItem):
        return item.opts['meshdata'].vertexes()
    elif isinstance(item, opengl.GLLinePlotItem):
        return item.pos
    elif isinstance(item, opengl.GLBarGraphItem):
        return item.points()
    return None

def compute_scene_bbox(items):
    min_vals, max_vals = None, None
    
    for item in items:
        data = get_item_data(item)
        if data is None or len(data) == 0:
            continue
            
        if not isinstance(data, np.ndarray):
            data = np.array(data)
            
        if data.ndim == 1:
            data = data[np.newaxis, :]
        if data.shape[1] == 2:
            data = np.hstack([data, np.zeros((len(data), 1))])
            
        item_min = np.nanmin(data, axis=0)
        item_max = np.nanmax(data, axis=0)
        
        if min_vals is None:
            min_vals, max_vals = item_min, item_max
        else:
            min_vals = np.minimum(min_vals, item_min)
            max_vals = np.maximum(max_vals, item_max)
    
    if min_vals is None:
        return np.array([0, 0, 0]), np.array([1, 1, 1])
    
    return min_vals, max_vals


class WeakList(object):

    def __init__(self):
        self._items = []

    def append(self, obj):
        self._items.insert(0, weakref.ref(obj))

    def __iter__(self):
        i = len(self._items)-1
        while i >= 0:
            ref = self._items[i]
            d = ref()
            if d is None:
                del self._items[i]
            else:
                yield d
            i -= 1

class CustomTransform(QtGui.QTransform):
    def __init__(self, parent=None):
        super().__init__()
        self._vertical_shear = 0.0 
        self._horizontal_shear = 0.0
        self.updateTransform()

    def getVerticalShear(self):
        return self._vertical_shear
    
    def setVerticalShear(self, shear):
        if self._vertical_shear != shear:
            self._vertical_shear = shear
            self.updateTransform()

    def getHorizontalShear(self):
        return self._horizontal_shear

    def setHorizontalShear(self, shear):
        if self._horizontal_shear != shear:
            self._horizontal_shear = shear
            self.updateTransform()

    def updateTransform(self):
        self.setMatrix(1.0, self._horizontal_shear, 0.0,
                      self._vertical_shear, 1.0, 0.0,
                      0.0, 0.0, 1.0)

    verticalShear = QtCore.Property(float, getVerticalShear, setVerticalShear)
    horizontalShear = QtCore.Property(float, getHorizontalShear, setHorizontalShear)


class CustomGraphicsItem(QtWidgets.QGraphicsItem):
    def __init__(self, parent=None):
        super().__init__(parent)



class CustomItemGroup(GLGraphicsItem.GLGraphicsItem):
    def __init__(self, parent=None):
        super().__init__(parent)
        QtWidgets.QGraphicsObject.setFlag(self, QtWidgets.QGraphicsObject.GraphicsItemFlag.ItemHasNoContents)
        
    def addItem(self, item):
        item.setParentItem(self)
        print("\n")
        print(item)
        print(item.parentItem())
        print(self)
        #self.normaliseToparent()

    def boundingRect(self):
        return self.childrenBoundingRect()
    
    def normaliseToparent(self):
        rect = self.boundingRect()
        
        tr = self.transform()
        tr.scale(1/rect.width(), 1/rect.height())
        tr.translate(-rect.x(), -rect.y())
        self.update()

class ChildGroup(CustomItemGroup):
    def __init__(self, parent):
        super().__init__(parent=parent)
        print(f"self.group(): {self.group()}")
        self.itemsChangedListeners = WeakList()

    def itemChange(self, change, value):
        ret = super().itemChange(change, value)
        if change in [
            QtWidgets.QGraphicsItem.ItemChildAddedChange,
            QtWidgets.QGraphicsItem.ItemChildRemovedChange,
        ]:
            try:
                for listener in self.itemsChangedListeners:
                    listener.itemsChanged()
            except AttributeError:
                pass
        return ret
    
    
class ViewBox(QtWidgets.QGraphicsItem):
    
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.__view=None

        #self.childGroup: ChildGroup = ChildGroup(parent=self)
        #self.childGroup.setParentItem(self)
        #self.childGroup.itemsChangedListeners.append(self)
        
    def _setView(self, v):
        self.__view = v
        
        
    def addItem(self, item):
        print(f"ViewBox addItem: {item}")
        self.group().addToGroup(item)
        #self.childGroup.addItem(item)
        #self.normalize_scene()
        
    def itemsChanged(self):
        print("itemsChanged")
        self.update()
    
    
    
    
    def normalize_scene(self):
        min_vals, max_vals = compute_scene_bbox(self.childItems())
        self.min_vals = min_vals
        self.max_vals = max_vals
        
        diff = max_vals - min_vals
        

        self.vb_min = min_vals - diff * self.padding
        self.vb_max = max_vals + diff* self.padding
        
        ranges = self.vb_max - self.vb_min
        ranges[ranges == 0] = 1.0
        
        self.transform_matrix = QtGui.QMatrix4x4()
        #tr = CustomTransform()
       # self.childGroup.setTransform(tr)

        self.transform_matrix.scale(1/ranges[0], 1/ranges[1], 1/ranges[2])
        self.transform_matrix.translate(-self.vb_min[0], -self.vb_min[1], -self.vb_min[2])
        for child in self.childItems():
            child.setTransform(self.transform_matrix)
            
    def boundingRect(self):
        return self.childrenBoundingRect()

class CustomGLViewWidget(opengl.GLViewWidget):  
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def addItem(self, item, ignoreBounds=False):
        if isinstance(item, ViewBox):
            self.setViewBox(item)
            return
        if not ignoreBounds and not isinstance(item, ViewBox):
            self.vb.addItem(item)
        super().addItem(item)   

    def setViewBox(self, vb):
        self.vb=vb


class Window(QtWidgets.QMainWindow):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.setWindowTitle("ItemGroup with QOpenGLWidget Example")
        
        
        self.gl_view = opengl.GLViewWidget()
        self.setCentralWidget(self.gl_view)
        self.view_box = ViewBox(parent=self.gl_view)
        
        self.gl_view.addItem(self.view_box)
        self.view_box.setGroup(QtWidgets.QGraphicsItemGroup())
        
        
        x_range, y_range = [-5, 5], [-5, 5]
        self.scatter, self.surface = self.create_surface(x_range, y_range)
        self.view_box.setParentItem(self.scatter)
        self.view_box.setParentItem(self.surface)
        #self.gl_view.addItem(self.scatter)
        #self.gl_view.addItem(self.surface)
        
        self.gl_view.opts["distance"]=25

        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.update_data)
        
        nticks=15
        for i, j in zip(np.linspace(*x_range, nticks), np.linspace(*y_range, nticks)):
            x = opengl.GLTextItem(pos = (i, 0, 0), text=f"{round(i, 1)}", color="white")
            y = opengl.GLTextItem(pos = (0, j, 0), text=f"{round(j, 1)}", color="yellow")
            z = opengl.GLTextItem(pos = (0, 0, i), text=f"{round(i, 1)}", color="cyan")
            self.gl_view.addItem(x, True)
            self.gl_view.addItem(y, True)
            self.gl_view.addItem(z, True)
        self.showMaximized()
        self.update()
        
        

        
       # self.timer.start(1500)
        
    def update_data(self):
        self.rect1.setRect(2 * 1200, 2*1200, 50, 50)

    def create_surface(self, x_range, y_range):
        X = np.arange(*x_range, 0.5)
        Y = np.arange(*y_range, 0.5)
        X_mat, Y_mat = np.meshgrid(X, Y)
        
        R_shifted = np.sqrt((X_mat)**2 + (Y_mat)**2)
        Z_mat = 2 + np.sin(R_shifted)                
        X_vec, Y_vec, Z_vect = X_mat.flatten(), Y_mat.flatten(), Z_mat.flatten()
        
        pos_scatter = np.column_stack((X_vec, Y_vec, Z_vect))
        pos_surface = [X, Y, 3*Z_mat]

        
        scatter = GLScatterPlotItem(pos=pos_scatter, color=(1, 0, 0, 1))
        surface = opengl.GLSurfacePlotItem(*pos_surface,
                                    glOptions='opaque',
                                    color=(0.5, 0.5, 1, 1))
        
        from types import MethodType
        return scatter, surface

def main():
    app = QtWidgets.QApplication(sys.argv)
    window = Window()
    sys.exit(app.exec())


if __name__ == "__main__":
    main()
    

    
